# mypy: disable-error-code=misc
import asyncio
import sys
import time
from http import HTTPStatus
from typing import Any, Optional

import langfuse
import structlog
from sanic import Blueprint, HTTPResponse, response
from sanic.request import Request
from sanic_openapi import openapi

import rasa
from rasa.builder.auth import (
    HEADER_USER_ID,
    email_from_auth,
    is_auth_required_now,
    protected,
)
from rasa.builder.config import (
    COPILOT_ASSISTANT_TRACKER_MAX_TURNS,
    COPILOT_HANDLER_ROLLING_BUFFER_SIZE,
    DEFAULT_BOT_BUILDER_EMAIL,
    GUARDRAILS_ENABLE_BLOCKING,
    HELLO_RASA_PROJECT_ID,
    LAKERA_ASSISTANT_HISTORY_GUARDRAIL_PROJECT_ID,
    LAKERA_COPILOT_HISTORY_GUARDRAIL_PROJECT_ID,
)
from rasa.builder.copilot.constants import DEFAULT_COPILOT_CHAT_ID, ROLE_USER
from rasa.builder.copilot.copilot_response_handler import CopilotResponseHandler
from rasa.builder.copilot.exceptions import CopilotStreamError
from rasa.builder.copilot.history_store import persist_copilot_message_to_history
from rasa.builder.copilot.models import (
    ConversationKey,
    CopilotChatMessage,
    CopilotContext,
    CopilotHistoryResponse,
    CopilotTurnRequest,
    GeneratedContent,
    ReferenceEntry,
    ReferenceSection,
    ResponseCategory,
    ResponseCompleteness,
    TextContent,
)
from rasa.builder.download import create_bot_project_archive
from rasa.builder.guardrails.constants import (
    BLOCK_SCOPE_PROJECT,
    BLOCK_SCOPE_USER,
    BlockScope,
)
from rasa.builder.guardrails.store import guardrails_store
from rasa.builder.job_manager import job_manager
from rasa.builder.jobs import (
    run_backup_to_bot_job,
    run_change_branch_job,
    run_prompt_to_bot_job,
    run_replace_all_files_job,
    run_revert_job,
    run_rollback_job,
    run_template_to_bot_job,
)
from rasa.builder.llm_service import llm_service
from rasa.builder.logging_utils import (
    capture_exception_with_context,
    get_recent_logs,
)
from rasa.builder.models import (
    AgentStatus,
    ApiErrorResponse,
    AssistantInfo,
    BotData,
    BotFiles,
    ChangeBranchRequest,
    GitCommitInfo,
    GitStatusResponse,
    JobCreateResponse,
    JobStatus,
    JobStatusEvent,
    PromptRequest,
    RestoreFromBackupRequest,
    ServerSentEvent,
    TemplateRequest,
)
from rasa.builder.project_generator import ProjectGenerator
from rasa.builder.shared.tracker_context import TrackerContext
from rasa.builder.telemetry.copilot_langfuse_telemetry import CopilotLangfuseTelemetry
from rasa.builder.telemetry.copilot_segment_telemetry import CopilotSegmentTelemetry
from rasa.core.agent import Agent
from rasa.core.channels.studio_chat import StudioChatInput
from rasa.core.exceptions import AgentNotReady
from rasa.shared.core.flows.flows_list import FlowsList
from rasa.shared.core.flows.yaml_flows_io import get_flows_as_json
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.importers.utils import DOMAIN_KEYS
from rasa.utils.json_utils import extract_values
from rasa.utils.openapi import model_to_schema

structlogger = structlog.get_logger()

# Create the blueprint
bp = Blueprint("bot_builder", url_prefix="/api")


def setup_project_generator(project_folder: str) -> ProjectGenerator:
    """Initialize and return a ProjectGenerator instance."""
    # Ensure the project folder is in sys.path
    if project_folder not in sys.path:
        sys.path.insert(0, project_folder)

    structlogger.info(
        "bot_builder_service.service_initialized", project_folder=project_folder
    )

    return ProjectGenerator(project_folder)


def get_project_generator(request: Request) -> ProjectGenerator:
    """Get the project generator from app context."""
    return request.app.ctx.project_generator


def get_input_channel(request: Request) -> StudioChatInput:
    """Get the input channel from app context."""
    return request.app.ctx.input_channel


async def extract_bot_data_from_agent(agent: Agent) -> BotData:
    """Extract BotData from an Agent.

    Args:
        agent: The agent to extract data from

    Returns:
        BotData containing flows, domain, config, endpoints, and nlu data
    """
    domain = agent.domain.as_dict() if agent.domain else {}
    flows = (
        await agent.processor.get_flows()
        if agent.processor
        else FlowsList(underlying_flows=[])
    )
    return BotData(
        flows=get_flows_as_json(flows),
        domain=extract_values(domain, DOMAIN_KEYS),
    )


async def get_agent_status(request: Request) -> AgentStatus:
    """Get the status of the agent."""
    if request.app.ctx.agent is None:
        return AgentStatus.not_loaded
    agent: Agent = request.app.ctx.agent
    if agent.is_ready():
        return AgentStatus.ready
    return AgentStatus.not_ready


# Health check endpoint
@bp.route("/", methods=["GET"])
@openapi.summary("Health check endpoint")
@openapi.description(
    "Returns the health status of the Bot Builder service including version "
    "information and authentication requirements"
)
@openapi.tag("health")
@openapi.response(
    200,
    {
        "application/json": {
            "status": str,
            "service": str,
            "rasa_version": str,
            "auth_required": bool,
            "agent_status": str,
        }
    },
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
async def health(request: Request) -> HTTPResponse:
    """Health check endpoint."""
    project_generator = get_project_generator(request)
    return response.json(
        {
            "status": "ok",
            "service": "bot-builder",
            "rasa_version": rasa.__version__,
            "agent_status": await get_agent_status(request),
            "auth_required": is_auth_required_now(
                project_info=project_generator.project_info
            ),
        }
    )


@bp.route("/job-events/<job_id>", methods=["GET"])
@openapi.summary("Stream job progress events")
@openapi.description(
    "Stream server-sent events (SSE) tracking real-time job progress.\n\n"
    "**Connect with:** `Accept: text/event-stream`.\n\n"
    "**SSE Event Example:**\n"
    "```text\n"
    "event: received\n"
    'data: {"status": "received"}\n'
    "```\n\n"
)
@openapi.tag("job-events")
@openapi.parameter(
    "job_id", str, location="path", description="The id of the job to stream events for"
)
@openapi.response(
    200,
    {"text/event-stream": str},
    description="Server-sent events stream. See documentation for event types.",
    example=(
        "event: received\n"
        'data: {"status": "received"}\n'
        "\n"
        "event: generating\n"
        'data: {"status": "generating"}\n'
    ),
)
@openapi.response(
    404,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Unknown job_id: No such job exists.",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def job_events(request: Request, job_id: str) -> HTTPResponse:
    try:
        job = job_manager.get_job(job_id)
        if job is None:
            return response.json(
                ApiErrorResponse(
                    error="Job not found", details={"job_id": job_id}
                ).model_dump(),
                status=404,
            )

        stream = await request.respond(content_type="text/event-stream")

        try:
            async for evt in job.event_stream():
                await stream.send(evt.format())
        except Exception as exc:
            # Handle exceptions within the SSE stream context
            capture_exception_with_context(
                exc,
                "bot_builder_service.job_events.streaming_error",
                extra={"job_id": job_id},
                tags={"endpoint": "/api/job-events/<job_id>"},
            )
            # Send error event in SSE format instead of JSON response
            error_event = JobStatusEvent.from_status(
                status=JobStatus.error.value,
                message=f"Failed to stream job events: {exc}",
            ).format()
            await stream.send(error_event)
        finally:
            await stream.eof()

        return stream
    except Exception as exc:
        # This exception handler only applies before stream.respond() is called
        capture_exception_with_context(
            exc,
            "bot_builder_service.job_events.unexpected_error",
            extra={"job_id": job_id},
            tags={"endpoint": "/api/job-events/<job_id>"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to stream job events", details={"error": str(exc)}
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/prompt-to-bot", methods=["POST"])
@openapi.summary("Generate bot from natural language prompt")
@openapi.description(
    "Creates a complete conversational AI bot from a natural language prompt. "
    "Returns immediately with a job ID. Connect to `/job-events/<job_id>` to "
    "receive server-sent events (SSE) for real-time progress tracking "
    "throughout the bot creation process.\n\n"
    "**SSE Event Flow** (via `/job-events/<job_id>`):\n"
    "1. `received` - Request received by server\n"
    "2. `generating` - Generating bot project files\n"
    "3. `generation_success` - Bot generation completed successfully\n"
    "4. `training` - Training the bot model\n"
    "5. `train_success` - Model training completed\n"
    "6. `done` - Bot creation completed\n\n"
    "**Error Events:**\n"
    "- `generation_error` - Failed to generate bot from prompt\n"
    "- `train_error` - Bot generated but training failed\n"
    "- `validation_error` - Generated bot configuration is invalid\n"
    "- `error` - Unexpected error occurred\n\n"
    "**Usage:**\n"
    "1. Send POST request with Content-Type: application/json\n"
    "2. The response will be a JSON object `{job_id: ...}`\n"
    "3. Connect to `/job-events/<job_id>` for a server-sent event stream of progress."
)
@openapi.tag("bot-generation")
@openapi.body(
    {"application/json": model_to_schema(PromptRequest)},
    description="Prompt request with natural language description.",
    required=True,
    example={
        "prompt": (
            "Create a customer support bot that can help users with order inquiries, "
            "product questions, and returns processing. The bot should be friendly "
            "and able to escalate to human agents when needed."
        )
    },
)
@openapi.response(
    200,
    {"application/json": model_to_schema(JobCreateResponse)},
    description="Job created. Poll or subscribe to /job-events/<job_id> for progress.",
)
@openapi.response(
    400,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Validation error in request payload",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def handle_prompt_to_bot(request: Request) -> HTTPResponse:
    """Handle prompt-to-bot generation requests."""
    try:
        payload = PromptRequest(**request.json)
    except Exception as exc:
        return response.json(
            ApiErrorResponse(
                error="Invalid request", details={"error": str(exc)}
            ).model_dump(),
            status=400,
        )

    try:
        # Allocate job and schedule background task
        job = job_manager.create_job()
        request.app.add_task(run_prompt_to_bot_job(request.app, job, payload.prompt))
        return response.json(JobCreateResponse(job_id=job.id).model_dump(), status=200)
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.prompt_to_bot.unexpected_error",
            tags={"endpoint": "/api/prompt-to-bot"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to create prompt-to-bot job",
                details={"error": str(exc)},
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/template-to-bot", methods=["POST"])
@openapi.summary("Generate bot from predefined template")
@openapi.description(
    "Creates a complete conversational AI bot from a predefined template. "
    "Returns immediately with a job ID. Connect to `/job-events/<job_id>` to "
    "receive server-sent events (SSE) for real-time progress tracking "
    "throughout the bot creation process.\n\n"
    "**SSE Event Flow** (via `/job-events/<job_id>`):\n"
    "1. `received` - Request received by server\n"
    "2. `generating` - Initializing bot from template\n"
    "3. `generation_success` - Template initialization completed successfully\n"
    "4. `training` - Training the bot model\n"
    "5. `train_success` - Model training completed\n"
    "6. `done` - Bot creation completed\n\n"
    "**Error Events:**\n"
    "- `generation_error` - Failed to initialize bot from template\n"
    "- `train_error` - Template loaded but training failed\n"
    "- `validation_error` - Template configuration is invalid\n"
    "- `error` - Unexpected error occurred\n\n"
    "**Usage:**\n"
    "1. Send POST request with Content-Type: application/json\n"
    "2. The response will be a JSON object `{job_id: ...}`\n"
    "3. Connect to `/job-events/<job_id>` for a server-sent event stream of progress."
)
@openapi.tag("bot-generation")
@openapi.body(
    {"application/json": model_to_schema(TemplateRequest)},
    description="Template request with template name.",
    required=True,
    example={"template_name": "telco"},
)
@openapi.response(
    200,
    {"application/json": model_to_schema(JobCreateResponse)},
    description="Job created. Poll or subscribe to /job-events/<job_id> for progress.",
)
@openapi.response(
    400,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Validation error in request payload or invalid template name",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def handle_template_to_bot(request: Request) -> HTTPResponse:
    """Create a new template-to-bot job and return job_id immediately."""
    try:
        template_data = TemplateRequest(**request.json)
    except Exception as exc:
        return response.json(
            ApiErrorResponse(
                error="Invalid request", details={"error": str(exc)}
            ).model_dump(),
            status=400,
        )

    try:
        # allocate job and schedule background task
        job = job_manager.create_job()
        request.app.add_task(
            run_template_to_bot_job(request.app, job, template_data.template_name)
        )
        return response.json(JobCreateResponse(job_id=job.id).model_dump(), status=200)
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.template_to_bot.unexpected_error",
            tags={"endpoint": "/api/template-to-bot"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to create template-to-bot job",
                details={"error": str(exc)},
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/backup-to-bot", methods=["POST"])
@openapi.summary("Generate bot from backup archive")
@openapi.description(
    "Creates a complete conversational AI bot from a backup tar.gz archive. "
    "Returns immediately with a job ID. Connect to `/job-events/<job_id>` to "
    "receive server-sent events (SSE) for real-time progress tracking "
    "throughout the bot restoration process.\n\n"
    "**SSE Event Flow** (via `/job-events/<job_id>`):\n"
    "1. `received` - Request received by server\n"
    "2. `generating` - Extracting and restoring bot from backup\n"
    "3. `generation_success` - Backup restoration completed successfully\n"
    "4. `training` - Training the bot model (if no existing model found)\n"
    "5. `train_success` - Model training completed (if training was needed)\n"
    "6. `done` - Bot restoration completed\n\n"
    "**Error Events:**\n"
    "- `generation_error` - Failed to restore bot from backup\n"
    "- `train_error` - Backup restored but training failed\n"
    "- `validation_error` - Restored bot configuration is invalid\n"
    "- `error` - Unexpected error occurred\n\n"
    "**Usage:**\n"
    "1. Send POST request with Content-Type: application/json\n"
    "2. The response will be a JSON object `{job_id: ...}`\n"
    "3. Connect to `/job-events/<job_id>` for a server-sent event stream of progress."
)
@openapi.tag("bot-generation")
@openapi.body(
    {"application/json": model_to_schema(RestoreFromBackupRequest)},
    description="Backup request with presigned URL to tar.gz archive.",
    required=True,
    example={"presigned_url": "https://s3.amazonaws.com/bucket/path?signature=..."},
)
@openapi.response(
    200,
    {"application/json": model_to_schema(JobCreateResponse)},
    description="Job created. Poll or subscribe to /job-events/<job_id> for progress.",
)
@openapi.response(
    400,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Validation error in request payload or invalid presigned URL",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def handle_backup_to_bot(request: Request) -> HTTPResponse:
    """Handle backup-to-bot restoration requests."""
    try:
        payload = RestoreFromBackupRequest(**request.json)
    except Exception as exc:
        return response.json(
            ApiErrorResponse(
                error="Invalid request", details={"error": str(exc)}
            ).model_dump(),
            status=400,
        )

    try:
        # Allocate job and schedule background task
        job = job_manager.create_job()
        request.app.add_task(
            run_backup_to_bot_job(request.app, job, payload.presigned_url)
        )
        return response.json(JobCreateResponse(job_id=job.id).model_dump(), status=200)
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.backup_to_bot.unexpected_error",
            tags={"endpoint": "/api/backup-to-bot"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to create backup-to-bot job",
                details={"error": str(exc)},
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/files", methods=["GET"])
@openapi.summary("Get bot files")
@openapi.description(
    "Retrieves the current bot configuration files including domain.yml, "
    "config.yml, flows.yml, NLU data, and other project files as a "
    "dictionary mapping file names to their string contents"
)
@openapi.tag("bot-files")
@openapi.response(
    200,
    {"application/json": {str: Optional[str]}},
    description="Bot files retrieved successfully",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def get_bot_files(request: Request) -> HTTPResponse:
    """Get current bot files."""
    try:
        project_generator = get_project_generator(request)
        bot_files = project_generator.get_bot_files()
        return response.json(bot_files)
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.get_bot_files.unexpected_error",
            tags={"endpoint": "/api/files", "method": "GET"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to retrieve bot files", details={"error": str(exc)}
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/files", methods=["POST"])
@openapi.summary("Replace all bot files")
@openapi.description(
    "Replaces all bot configuration files with the provided files, deletes any "
    "files not included in the request (excluding .rasa/ and models/ directories), "
    "and retrains the model. Returns immediately with a job ID. Connect to "
    "`/job-events/<job_id>` for real-time SSE progress tracking."
    "\n\n"
    "**File Management:**\n"
    "- All files in the request are written to the project folder\n"
    "- Files not included in the request are deleted from the project\n"
    "- Files/folders starting with `.rasa/` or `models/` are excluded from deletion\n\n"
    "**SSE Event Flow:** (available via /job-events/<job_id>)\n"
    "1. `received` - Request received by server\n"
    "2. `validating` - Validating bot configuration files\n"
    "3. `validation_success` - File validation completed successfully\n"
    "4. `training` - Training the bot model with updated files\n"
    "5. `train_success` - Model training completed\n"
    "6. `done` - Bot files replacement completed\n\n"
    "**Error Events (can occur at any time):**\n"
    "- `validation_error` - Bot configuration files are invalid\n"
    "- `train_error` - Files updated but training failed\n"
    "- `copilot_analysis_start` - Copilot analysis started (includes `copilot_job_id` "
    "in payload)\n"
    "- `error` - Unexpected error occurred\n\n"
    "**Copilot Analysis Events:**\n"
    "When training or validation fails, a separate copilot analysis job is "
    "automatically started. The `copilot_analysis_start` event includes a "
    "`copilot_job_id` in the payload.\nConnect to `/job-events/<copilot_job_id>` to "
    "receive the following events:\n"
    "- `copilot_analyzing` - Copilot is analyzing errors and providing suggestions. "
    "Uses the same SSE event payload format as the `/copilot` endpoint with `content`, "
    "`response_category`, and `completeness` fields.\n"
    "- `copilot_analysis_success` - Copilot analysis completed with references.\n"
    "- `copilot_analysis_error` - Copilot analysis failed\n\n"
    "**Usage:**\n"
    "1. Send POST request with Content-Type: application/json\n"
    "2. The response will be a JSON object `{job_id: ...}`\n"
    "3. Connect to `/job-events/<job_id>` for a server-sent event stream of progress."
)
@openapi.tag("bot-files")
@openapi.body(
    {"application/json": {str: Optional[str]}},
    description=(
        "A dictionary mapping file names to their complete content. "
        "All files in the project will be replaced with these files. "
        "Files not included in the request will be deleted from the project "
        "(except for .rasa/ and models/ directories). "
        "The file name should be the relative path from the project root."
    ),
    required=True,
    example={
        "domain.yml": (
            "version: '3.1'\n"
            "intents:\n  - greet\n  - goodbye\n"
            "responses:\n  utter_greet:\n  - text: 'Hello!'\n"
            "  utter_goodbye:\n  - text: 'Goodbye!'"
        ),
        "config.yml": (
            "version: '3.1'\n"
            "pipeline:\n  - name: WhitespaceTokenizer\n"
            "  - name: RegexFeaturizer\n  - name: LexicalSyntacticFeaturizer\n"
            "  - name: CountVectorsFeaturizer\n"
            "policies:\n  - name: MemoizationPolicy\n  - name: RulePolicy"
        ),
        "data/nlu.yml": (
            "version: '3.1'\n"
            "nlu:\n- intent: greet\n  examples: |\n    - hello\n    - hi"
        ),
    },
)
@openapi.response(
    200,
    {"application/json": model_to_schema(JobCreateResponse)},
    description=(
        "Job created. Poll or subscribe to /job-events/<job_id> "
        "for progress and SSE updates on file replacement and training."
    ),
)
@openapi.response(
    400,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Validation error in bot files",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.response(
    401,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description=(
        "Authentication failed - Authorization header missing or invalid. "
        "Auth may be conditionally required depending on server "
        "configuration."
    ),
)
@openapi.parameter(
    "Authorization",
    description=(
        "Bearer token for authentication. Required after auth start window "
        "or if configured."
    ),
    _in="header",
    required=False,
    schema=str,
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
@openapi.parameter(
    "X-Commit-Message",
    description=(
        "Optional custom commit message. If not provided, AI-generated message used."
    ),
    _in="header",
    required=False,
    schema=str,
)
@protected()
async def replace_all_bot_files(request: Request) -> HTTPResponse:
    """Replace all bot files with server-sent events for progress tracking."""
    try:
        bot_files = request.json
    except Exception as exc:
        return response.json(
            ApiErrorResponse(
                error="Invalid request", details={"error": str(exc)}
            ).model_dump(),
            status=400,
        )

    try:
        # Extract commit metadata from headers
        commit_message = request.headers.get("X-Commit-Message")
        email = email_from_auth(request) or DEFAULT_BOT_BUILDER_EMAIL
        commit_info = GitCommitInfo(
            author="Bot Builder", email=email, message=commit_message
        )

        job = job_manager.create_job()
        request.app.add_task(
            run_replace_all_files_job(request.app, job, bot_files, commit_info)
        )
        return response.json(JobCreateResponse(job_id=job.id).model_dump(), status=200)
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.replace_all_bot_files.unexpected_error",
            tags={"endpoint": "/api/files", "method": "POST"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to replace bot files", details={"error": str(exc)}
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/data", methods=["GET"])
@openapi.summary("Get bot data")
@openapi.description(
    "Retrieves the current bot data in CALM import format with flows, domain, "
    "config, endpoints, and NLU data"
)
@openapi.tag("bot-info")
@openapi.response(
    200,
    {"application/json": model_to_schema(BotData)},
    description="Bot data retrieved successfully",
)
@openapi.response(
    409,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Agent not ready",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def get_bot_data(request: Request) -> HTTPResponse:
    """Get current bot data in CALM import format."""
    try:
        agent: Optional[Agent] = request.app.ctx.agent
        if not agent:
            raise AgentNotReady(
                "Can't retrieve the data without an agent being loaded."
            )

        bot_data = await extract_bot_data_from_agent(agent)

        return response.json(bot_data.model_dump())
    except AgentNotReady as e:
        return response.json(
            ApiErrorResponse(
                error="Agent not ready",
                details={"error": str(e)},
            ).model_dump(),
            status=HTTPStatus.CONFLICT,
        )
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.get_bot_data.unexpected_error",
            tags={"endpoint": "/api/data"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to retrieve bot data",
                details={"error": str(exc)},
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/assistant", methods=["GET"])
@openapi.summary("Get assistant info")
@openapi.description(
    "Returns basic information about the loaded assistant, including the assistant id "
    "as configured in the model's metadata (from config.yml)."
)
@openapi.tag("bot-info")
@openapi.response(
    200,
    {"application/json": model_to_schema(AssistantInfo)},
    description="Assistant info retrieved successfully",
)
@openapi.response(
    409,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Agent not ready",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def get_bot_info(request: Request) -> HTTPResponse:
    """Return assistant info including assistant id from model metadata."""
    try:
        agent: Optional[Agent] = request.app.ctx.agent
        if not agent:
            raise AgentNotReady("Can't retrieve bot info without a loaded agent.")

        assistant_id: Optional[str] = (
            agent.processor.model_metadata.assistant_id
            if agent.processor
            and agent.processor.model_metadata
            and hasattr(agent.processor.model_metadata, "assistant_id")
            else None
        )

        return response.json(AssistantInfo(assistant_id=assistant_id).model_dump())
    except AgentNotReady as e:
        return response.json(
            ApiErrorResponse(
                error="Agent not ready",
                details={"error": str(e)},
            ).model_dump(),
            status=HTTPStatus.CONFLICT,
        )
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.get_bot_info.unexpected_error",
            tags={"endpoint": "/api/assistant"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to retrieve bot info",
                details={"error": str(exc)},
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/download", methods=["GET"])
@openapi.summary("Download bot project as tar.gz")
@openapi.description(
    "Downloads the current bot project files as a compressed tar.gz archive. "
    "Includes all configuration files and a .env file with RASA_PRO_LICENSE."
)
@openapi.tag("bot-files")
@openapi.parameter(
    "exclude_models_directory",
    bool,
    location="query",
    description="Whether to exclude the models directory",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
@openapi.parameter(
    "project_name",
    description="Name of the project for the archive filename and pyproject.toml",
    _in="query",
    required=False,
    schema=str,
)
@openapi.response(
    200,
    {"application/gzip": bytes},
    description="Bot project downloaded successfully as tar.gz",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
async def download_bot_project(request: Request) -> HTTPResponse:
    """Download bot project as tar.gz archive."""
    try:
        # Get query parameters
        exclude_models_directory = (
            request.args.get("exclude_models_directory", "true").lower() == "true"
        )
        project_name = request.args.get("project_name", "bot-project")

        # Get bot files
        project_generator = get_project_generator(request)
        bot_files = project_generator.get_bot_files(
            exclude_models_directory=exclude_models_directory
        )

        # Create tar.gz archive
        tar_data = create_bot_project_archive(
            bot_files, project_name, project_generator.project_folder
        )

        structlogger.info(
            "bot_builder_service.download_bot_project.success",
            user_sub=(getattr(request.ctx, "auth_payload", None) or {}).get("sub"),
            files_count=len(bot_files),
            archive_size=len(tar_data),
            payload=getattr(request.ctx, "auth_payload", None),
            project_name=project_name,
        )

        return response.raw(
            tar_data,
            content_type="application/gzip",
            headers={
                "Content-Disposition": f"attachment; filename={project_name}.tar.gz"
            },
        )

    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.download_bot_project.unexpected_error",
            tags={"endpoint": "/api/download"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to create bot project archive",
                details={"error": str(exc)},
            ).model_dump(),
            status=500,
        )


@bp.route("/copilot", methods=["POST"])
@openapi.summary("AI copilot for bot building")
@openapi.description(
    "Provides LLM-powered copilot assistance with streaming markdown responses. "
    "Returns server-sent events (SSE) for real-time streaming of copilot responses.\n\n"
    "The event's `event` field is the type of event. For this endpoint it will always "
    "be: `copilot_response`.\n\n"
)
@openapi.tag("copilot")
@openapi.body(
    {"application/json": model_to_schema(CopilotTurnRequest)},
    description=(
        "Copilot request containing: "
        "1. a single user message, "
        "2. session ID for tracking conversation context with the bot being built, "
        "3. optional chat ID for copilot conversation (defaults to 'default'), "
        "4. optional project ID for conversation isolation, "
        "Conversation history is stored and managed on the server."
    ),
    required=True,
)
@openapi.response(
    200,
    {"text/event-stream": model_to_schema(ServerSentEvent)},
    description=(
        "Server-sent events stream with copilot responses and references. "
        "The event's `event` field is the type "
        "of event. For this endpoint it will always be: `copilot_response`. "
        "The event's data field is a JSON dump. The following describes the event data "
        "field:\n"
        "- `response_category` can be one of the following:\n"
        "  - `copilot` - Stream token generated by the copilot.\n"
        "  - `out_of_scope_detection` - Response coming from the the out of scope detection.\n"  # noqa: E501
        "  - `roleplay_detection` - Response coming from the the roleplay detection.\n"
        "  - `reference` - Reference section.\n"
        "  - `training_error_log_analysis` - Used to flag the responses that are training error log analysis.\n"  # noqa: E501
        "  - `e2e_testing_error_log_analysis` - Used to flag the responses that are e2e testing error log analysis.\n"  # noqa: E501
        "  - `guardrail_policy_violation` - Used to flag the responses that are flagged as a violation of the guardrail policy.\n\n"  # noqa: E501
        "- `completeness`: Whether this is a streaming token or complete response.\n\n"
        "- `content`: The actual response content (for streaming tokens).\n\n"
        "- `references`: (Only for `reference` response category) List of reference entries. Each reference entry is a dictionary with the following keys:\n\n"  # noqa: E501
        "   - `index`: Reference index as an integer number.\n"
        "   - `title`: Reference title as a string.\n"
        "   - `url`: Reference URL as a string.\n\n"
    ),
    examples={
        "Streaming token from copilot": {
            "summary": "Streaming token from copilot",
            "value": GeneratedContent(
                content="<token generated by the copilot>",
                response_category=ResponseCategory.COPILOT,
                response_completeness=ResponseCompleteness.TOKEN,
            )
            .to_sse_event()
            .model_dump(),
        },
        "Complete response for the following response categories: out_of_scope_detection / roleplay_detection / training_error_log_analysis / e2e_testing_error_log_analysis / guardrail_policy_violation": {  # noqa: E501
            "summary": "Complete response from copilot",
            "value": GeneratedContent(
                content="<response from the out of scope detection>",
                response_category=ResponseCategory.OUT_OF_SCOPE_DETECTION,
                response_completeness=ResponseCompleteness.COMPLETE,
            )
            .to_sse_event()
            .model_dump(),
        },
        "Response with the references": {
            "summary": "Reference section with all references",
            "value": ReferenceSection(
                references=[
                    ReferenceEntry(
                        index=1,
                        title="Title of the reference",
                        url="https://rasa.com/docs/...",
                    ),
                    ReferenceEntry(
                        index=2,
                        title="Title of another reference",
                        url="https://rasa.com/docs/...",
                    ),
                ],
                response_category=ResponseCategory.REFERENCE,
                response_completeness=ResponseCompleteness.COMPLETE,
            )
            .to_sse_event()
            .model_dump(),
        },
    },
)
@openapi.response(
    400,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Validation error in request",
)
@openapi.response(
    502,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="LLM generation failed.",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error.",
)
@openapi.response(
    401,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description=(
        "Authentication failed - Authorization header missing or invalid. "
        "Auth may be conditionally required depending on server "
        "configuration."
    ),
)
@openapi.parameter(
    "Authorization",
    description=(
        "Bearer token for authentication. Required after auth start window "
        "or if configured."
    ),
    _in="header",
    required=False,
    schema=str,
)
@openapi.parameter(
    HEADER_USER_ID,
    description=("Required user id used for telemetry and guardrails (X-User-Id)."),
    _in="header",
    required=True,
    schema=str,
)
@protected()
# Disable automatic input/output capture for langfuse tracing
# This allows manual control over what data is sent to langfuse
@langfuse.observe(capture_input=False, capture_output=False)
async def copilot(request: Request) -> None:
    """Handle copilot requests with streaming markdown responses."""
    sse = await request.respond(content_type="text/event-stream")
    project_generator = get_project_generator(request)

    try:
        # 1. Validate and unpack input
        req = CopilotTurnRequest(**request.json)

        # Require user identifier via header and fail fast if missing
        user_id = (request.headers.get(HEADER_USER_ID) or "").strip()
        if not user_id:
            structlogger.error(
                "bot_builder_service.copilot.missing_or_empty_user_id_header",
                event_info=f"Missing or empty required header: {HEADER_USER_ID}",
            )
            await sse.send(
                ServerSentEvent(
                    event="error",
                    data={
                        "error": f"Missing or empty required header: {HEADER_USER_ID}"
                    },
                ).format()
            )
            return

        telemetry = CopilotSegmentTelemetry(
            project_id=HELLO_RASA_PROJECT_ID, user_id=user_id
        )
        structlogger.debug("builder.copilot.telemetry.request.init")

        # TODO: This can be removed once Langfuse is completed.
        if req.message and req.message.role == ROLE_USER:
            structlogger.debug("builder.copilot.telemetry.request.user_turn")
            # Offload telemetry logging to a background task
            request.app.add_task(
                asyncio.to_thread(
                    telemetry.log_user_turn,
                    req.message.get_flattened_text_content(),
                )
            )

        # 2. Check if we need to block the request due to too many guardrails violations
        if (scope := await _get_copilot_block_scope(user_id)) is not None:
            message = CopilotResponseHandler.respond_to_guardrail_blocked(scope)
            await sse.send(message.to_sse_event().format())
            return

        # 3. Load existing server-side history and build candidate turn
        # Use chat_id from payload if provided, otherwise use default
        chat_id = (req.chat_id or "").strip() or DEFAULT_COPILOT_CHAT_ID
        conversation_key = ConversationKey(chat_id=chat_id)
        existing_history = await llm_service.history_store.get(conversation_key)

        # Build the full history for context (existing + new user message)
        candidate_history = [*existing_history, req.message]

        # 4. Get the necessary context for the copilot
        tracker_context = await get_tracker_context_for_copilot(request, req, user_id)
        relevant_assistant_files = get_relevant_assistant_files_for_copilot(
            project_generator,
        )
        context = CopilotContext(
            tracker_context=tracker_context,
            assistant_logs=get_recent_logs(),
            assistant_files=relevant_assistant_files,
            copilot_chat_history=candidate_history,
        )

        # Persist the user message to history
        try:
            await llm_service.history_store.append(conversation_key, req.message)
        except Exception as exc:
            structlogger.error(
                "builder.copilot.history.user_message_persist_failed", error=str(exc)
            )

        # 5. Run guardrail policy checks. If any policy violations are detected,
        #    send a response and end the stream.
        guardrail_response: Optional[GeneratedContent] = None
        if llm_service.guardrails_policy_checker is not None:
            guardrail_response = await llm_service.guardrails_policy_checker.check_copilot_chat_for_policy_violations(  # noqa: E501
                context=context,
                hello_rasa_user_id=user_id,
                hello_rasa_project_id=HELLO_RASA_PROJECT_ID,
                lakera_project_id=LAKERA_COPILOT_HISTORY_GUARDRAIL_PROJECT_ID,
            )
        if guardrail_response is not None:
            blocked_or_violation_message = (
                await _handle_guardrail_violation_and_maybe_block(
                    sse=sse,
                    user_id=user_id,
                    violation_response=guardrail_response,
                )
            )

            # Persist the guardrail response as well
            guardrail_chat_message = CopilotChatMessage(
                role="copilot",
                content=[
                    TextContent(type="text", text=blocked_or_violation_message.content)
                ],
                response_category=blocked_or_violation_message.response_category,
            )
            try:
                await llm_service.history_store.append(
                    conversation_key, guardrail_chat_message
                )
            except Exception as exc:
                structlogger.error(
                    "builder.copilot.history.guardrail_response_persist_failed",
                    error=str(exc),
                )
            return

        # 6. Get the original response stream from copilot and handle it with the
        #    copilot response handler
        start_timestamp = time.perf_counter()
        copilot_client = llm_service.instantiate_copilot()
        (original_stream, generation_context) = await copilot_client.generate_response(
            context
        )

        copilot_response_handler = llm_service.instantiate_handler(
            COPILOT_HANDLER_ROLLING_BUFFER_SIZE
        )
        intercepted_stream = copilot_response_handler.handle_response(original_stream)

        # 7. Stream the intercepted response
        async for token in intercepted_stream:
            await sse.send(token.to_sse_event().format())

        # 8a. Offload metabase telemetry logging to a background task
        usage_stats = copilot_client.usage_statistics
        request.app.add_task(
            asyncio.to_thread(
                telemetry.log_copilot_from_handler,
                handler=copilot_response_handler,
                used_documents=generation_context.relevant_documents,
                latency_ms=int((time.perf_counter() - start_timestamp) * 1000),
                system_message=generation_context.system_message,
                chat_history=generation_context.chat_history,
                last_user_message=(
                    req.message.get_flattened_text_content()
                    if (req.message and req.message.role == ROLE_USER)
                    else None
                ),
                tracker_event_attachments=generation_context.tracker_event_attachments,
                model=usage_stats.model or "N/A",
                prompt_tokens=usage_stats.prompt_tokens or 0,
                completion_tokens=usage_stats.completion_tokens or 0,
                total_tokens=usage_stats.total_tokens or 0,
                cached_prompt_tokens=(
                    copilot_client.usage_statistics.cached_prompt_tokens or 0
                ),
            )
        )
        # 8b. Setup output trace attributes for Langfuse
        CopilotLangfuseTelemetry.setup_copilot_endpoint_call_trace_attributes(
            hello_rasa_project_id=HELLO_RASA_PROJECT_ID or "N/A",
            chat_id=req.session_id or "N/A",
            user_id=user_id,
            request=req,
            handler=copilot_response_handler,
            relevant_documents=generation_context.relevant_documents,
            copilot_context=context,
        )

        # 9. Once the stream is over, extract and send references
        #    if any documents were used
        reference_section = None
        if generation_context.relevant_documents:
            reference_section = copilot_response_handler.extract_references(
                generation_context.relevant_documents
            )
            await sse.send(reference_section.to_sse_event().format())

        # 10. Append final assistant message to server-side history
        full_text, category = copilot_response_handler.extract_full_text_and_category()
        if full_text:
            try:
                # Pass references directly if they exist
                references = reference_section.references if reference_section else None
                await persist_copilot_message_to_history(
                    text=full_text,
                    chat_id=chat_id,
                    response_category=category,
                    references=references,
                )
            except Exception as exc:
                structlogger.error(
                    "builder.copilot.history.persist_failed", error=str(exc)
                )
        else:
            structlogger.warning(
                "builder.copilot.history.no_assistant_text",
                session_id=req.session_id,
            )

    except CopilotStreamError as e:
        capture_exception_with_context(
            e,
            "bot_builder_service.copilot.generation_error",
            extra={"session_id": req.session_id},
            tags={"endpoint": "/api/copilot"},
        )
        await sse.send(
            ServerSentEvent(
                event="error",
                data={"error": str(e)},
            ).format()
        )

    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.copilot.unexpected_error",
            extra={"session_id": req.session_id if "req" in locals() else None},
            tags={"endpoint": "/api/copilot"},
        )
        await sse.send(
            ServerSentEvent(
                event="error",
                data={"error": str(exc)},
            ).format()
        )

    finally:
        await sse.eof()


@bp.route("/copilot/internal_message_templates/<template_name>", methods=["GET"])
@openapi.summary("Get templated response for copilot internal message")
@openapi.description(
    "Returns the templated response text for a given template name from the "
    "copilot internal message formatter. This endpoint provides access to the "
    "predefined templates used for formatting internal system messages."
)
@openapi.tag("copilot")
@openapi.parameter(
    "template_name",
    str,
    location="path",
    description=(
        "The template name to get the template for."
        "(e.g., 'training_error_log_analysis', 'e2e_testing_error_log_analysis')",
    ),
)
@openapi.response(
    200,
    {"application/json": {"template": str, "template_name": str}},
    description="Successfully retrieved the template for the given template name",
    example={
        "template": "The assistant training failed. Your task is ...",
        "template_name": "training_error_log_analysis",
    },
)
@openapi.response(
    404,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Template not found for the given template name",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def get_copilot_internal_message_template(
    request: Request, template_name: str
) -> HTTPResponse:
    """Get templated response for copilot internal message formatter."""
    try:
        # Try to get the template for the given template name
        template = llm_service.copilot_internal_message_templates.get(template_name)
        structlogger.info(
            "bot_builder_service.get_copilot_internal_message_template.template_found",
            template_name=template_name,
            template=template,
        )

        if template is None:
            structlogger.warning(
                "bot_builder_service.get_copilot_internal_message_template.not_found",
                template_name=template_name,
            )
            return response.json(
                ApiErrorResponse(
                    error="Template not found", details={"template_name": template_name}
                ).model_dump(),
                status=404,
            )

        return response.json({"template": template, "template_name": template_name})

    except Exception as e:
        structlogger.error(
            "bot_builder_service.get_copilot_internal_message_template.error",
            error=str(e),
            template_name=template_name,
        )
        return response.json(
            ApiErrorResponse(
                error="Internal server error", details={"template_name": template_name}
            ).model_dump(),
            status=500,
        )


@bp.route("/copilot/history", methods=["GET"])
@openapi.summary("Get Copilot chat history")
@openapi.description("Return stored copilot chat history.")
@openapi.tag("copilot")
@openapi.parameter(
    "chat_id",
    description="Optional chat ID to get history for (defaults to 'default').",
    _in="query",
    required=False,
    schema=str,
)
@openapi.response(
    200,
    {"application/json": model_to_schema(CopilotHistoryResponse)},
    description="Chat history retrieved successfully",
)
async def get_copilot_history(request: Request) -> HTTPResponse:
    """Get copilot chat history."""
    # Use chat_id from query parameter if provided, otherwise use default
    chat_id = (request.args.get("chat_id") or "").strip() or DEFAULT_COPILOT_CHAT_ID
    conversation_key = ConversationKey(chat_id=chat_id)
    messages = await llm_service.history_store.get(conversation_key)
    return response.json(CopilotHistoryResponse(messages=messages).model_dump())


@bp.route("/copilot/history", methods=["DELETE"])
@openapi.summary("Delete Copilot chat history")
@openapi.description("Delete stored copilot chat history.")
@openapi.tag("copilot")
@openapi.parameter(
    "chat_id",
    description="Optional chat ID to delete history for (defaults to 'default').",
    _in="query",
    required=False,
    schema=str,
)
@openapi.response(
    200,
    {
        "application/json": {
            "type": "object",
            "properties": {
                "status": {"type": "string"},
            },
        }
    },
    description="History deleted successfully",
)
async def delete_copilot_history(request: Request) -> HTTPResponse:
    """Delete copilot chat history."""
    # Use chat_id from query parameter if provided, otherwise use default
    chat_id = (request.args.get("chat_id") or "").strip() or DEFAULT_COPILOT_CHAT_ID
    conversation_key = ConversationKey(chat_id=chat_id)
    await llm_service.history_store.delete(conversation_key)
    return response.json({"status": "deleted"})


async def current_tracker_from_input_channel(
    app: Any, session_id: str
) -> Optional[DialogueStateTracker]:
    """Generate chat bot context from current conversation."""
    if app.ctx.agent and session_id:
        return await app.ctx.agent.tracker_store.retrieve(session_id)
    else:
        return None


async def _get_copilot_block_scope(user_id: str) -> Optional[BlockScope]:
    """Return the guardrail block scope for Copilot if blocked.

    Args:
        user_id: User identifier.

    Returns:
        'user' or 'project' when blocked, otherwise None.
    """
    if not GUARDRAILS_ENABLE_BLOCKING:
        return None

    return await guardrails_store.check_block_scope(user_id)


async def _handle_guardrail_violation_and_maybe_block(
    sse: Any,
    violation_response: GeneratedContent,
    user_id: str,
) -> GeneratedContent:
    """Record a violation, apply block if threshold crossed, and respond.

    Args:
        sse: Active SSE stream.
        violation_response: The default violation warning response.
        user_id: User identifier.

    Returns:
        The GeneratedContent message that was sent to the client.
    """
    if not GUARDRAILS_ENABLE_BLOCKING:
        await sse.send(violation_response.to_sse_event().format())
        return violation_response

    result = await guardrails_store.record_violation(user_id)

    if result.user_blocked_now:
        message = CopilotResponseHandler.respond_to_guardrail_blocked(BLOCK_SCOPE_USER)
    elif result.project_blocked_now:
        message = CopilotResponseHandler.respond_to_guardrail_blocked(
            BLOCK_SCOPE_PROJECT
        )
    else:
        message = violation_response

    await sse.send(message.to_sse_event().format())
    return message


@bp.route("/git/branch", methods=["POST"])
@openapi.summary("Change Git branch")
@openapi.description(
    "Changes the Git branch for the project folder and retrains the agent. "
    "Returns immediately with a job ID. Connect to `/job-events/<job_id>` "
    "to receive server-sent events (SSE) for real-time progress tracking.\n\n"
    "**SSE Event Flow** (via `/job-events/<job_id>`):\n"
    "1. `received` - Request received by server\n"
    "2. `switching_branch` - Switching to the specified branch\n"
    "3. `branch_switch_success` - Branch switched successfully\n"
    "4. `training` - Training the bot model with the new branch\n"
    "5. `train_success` - Model training completed\n"
    "6. `done` - Branch change completed\n\n"
    "**Error Events:**\n"
    "- `branch_switch_error` - Failed to switch to the branch\n"
    "- `train_error` - Branch switched but training failed\n"
    "- `validation_error` - Branch configuration is invalid\n"
    "- `error` - Unexpected error occurred\n\n"
    "**Usage:**\n"
    "1. Send POST request with Content-Type: application/json\n"
    "2. The response will be a JSON object `{job_id: ...}`\n"
    "3. Connect to `/job-events/<job_id>` for a server-sent event stream of progress."
)
@openapi.tag("git")
@openapi.body(
    {"application/json": model_to_schema(ChangeBranchRequest)},
    description="Branch change request with branch name and creation option.",
    required=True,
    example={"branch_name": "feature/new-flow", "create_if_not_exists": True},
)
@openapi.response(
    200,
    {"application/json": model_to_schema(JobCreateResponse)},
    description="Job created. Poll or subscribe to /job-events/<job_id> for progress.",
)
@openapi.response(
    400,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Validation error in request payload",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def handle_change_branch(request: Request) -> HTTPResponse:
    """Handle change branch requests."""
    try:
        payload = ChangeBranchRequest(**request.json)
    except Exception as exc:
        return response.json(
            ApiErrorResponse(
                error="Invalid request", details={"error": str(exc)}
            ).model_dump(),
            status=400,
        )

    try:
        # Allocate job and schedule background task
        job = job_manager.create_job()
        request.app.add_task(
            run_change_branch_job(
                request.app, job, payload.branch_name, payload.create_if_not_exists
            )
        )
        return response.json(JobCreateResponse(job_id=job.id).model_dump(), status=200)
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.change_branch.unexpected_error",
            tags={"endpoint": "/api/git/branch"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to create change branch job",
                details={"error": str(exc)},
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/git/status", methods=["GET"])
@openapi.summary("Get Git repository status")
@openapi.description(
    "Returns the current Git repository status including current branch, "
    "and uncommitted changes status."
)
@openapi.tag("git")
@openapi.response(
    200,
    {"application/json": model_to_schema(GitStatusResponse)},
    description="Git status retrieved successfully",
    example={
        "current_branch": "main",
        "uncommitted_changes": False,
    },
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def handle_git_status(request: Request) -> HTTPResponse:
    """Handle Git status requests."""
    try:
        project_generator = get_project_generator(request)
        git_service = project_generator.git_service

        git_status = GitStatusResponse(
            current_branch=await git_service.get_current_branch(),
            uncommitted_changes=await git_service.has_uncommitted_changes(),
        )

        return response.json(git_status.model_dump())
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.git_status.unexpected_error",
            tags={"endpoint": "/api/git/status"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to get Git status",
                details={"error": str(exc)},
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/commits", methods=["GET"])
@openapi.summary("Get commit history")
@openapi.description(
    "Returns the Git commit history for the project repository. "
    "Each commit includes SHA, author, timestamp, and message information."
)
@openapi.tag("git")
@openapi.parameter(
    "limit",
    description="Maximum number of commits to return (default: 50, max: 100)",
    _in="query",
    required=False,
    schema=int,
)
@openapi.response(
    200,
    {"application/json": {"commits": list}},
    description="Commit history retrieved successfully",
    example={
        "commits": [
            {
                "sha": "abc123def456",
                "short_sha": "abc123d",
                "author": "user",
                "email": "user@example.com",
                "timestamp": 1640995200,
                "message": "Update bot files",
            }
        ]
    },
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def handle_get_commits(request: Request) -> HTTPResponse:
    """Handle get commits requests."""
    try:
        project_generator = get_project_generator(request)
        limit = min(int(request.args.get("limit", 50)), 100)
        commits = await project_generator.git_service.get_commit_history(limit)
        return response.json({"commits": commits})
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.get_commits.unexpected_error",
            tags={"endpoint": "/api/commits"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to get commit history",
                details={"error": str(exc)},
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/commits/<commit_sha>/rollback", methods=["POST"])
@openapi.summary("Rollback to commit")
@openapi.description(
    "Rollback the project to a specific commit. This creates a background job "
    "that checks out the specified commit and retrains the agent. "
    "Returns immediately with a job ID. Connect to `/job-events/<job_id>` "
    "to receive server-sent events (SSE) for real-time progress tracking.\n\n"
    "**SSE Event Flow** (via `/job-events/<job_id>`):\n"
    "1. `received` - Request received by server\n"
    "2. `rolling_back` - Rolling back to the specified commit\n"
    "3. `rollback_success` - Commit checkout completed\n"
    "4. `training` - Training the bot model with the rolled back files\n"
    "5. `train_success` - Model training completed\n"
    "6. `done` - Rollback completed\n\n"
    "**Error Events:**\n"
    "- `rollback_error` - Failed to rollback to the commit\n"
    "- `train_error` - Rollback succeeded but training failed\n"
    "- `error` - Unexpected error occurred\n\n"
    "**Usage:**\n"
    "1. Send POST request\n"
    "2. The response will be a JSON object `{job_id: ...}`\n"
    "3. Connect to `/job-events/<job_id>` for a server-sent event stream of progress."
)
@openapi.tag("git")
@openapi.response(
    200,
    {"application/json": model_to_schema(JobCreateResponse)},
    description="Job created. Poll or subscribe to /job-events/<job_id> for progress.",
)
@openapi.response(
    400,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Invalid commit SHA",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def handle_rollback_to_commit(request: Request, commit_sha: str) -> HTTPResponse:
    """Handle rollback to commit requests."""
    try:
        # Validate commit SHA format
        if not commit_sha or len(commit_sha) < 7:
            return response.json(
                ApiErrorResponse(
                    error="Invalid commit SHA", details={"commit_sha": commit_sha}
                ).model_dump(),
                status=400,
            )

        job = job_manager.create_job()
        request.app.add_task(run_rollback_job(request.app, job, commit_sha))
        return response.json(JobCreateResponse(job_id=job.id).model_dump(), status=200)
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.rollback_to_commit.unexpected_error",
            tags={"endpoint": "/api/commits/<commit_sha>/rollback"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to create rollback job",
                details={"error": str(exc)},
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/commits/<commit_sha>/diff", methods=["GET"])
@openapi.summary("Get commit diff")
@openapi.description(
    "Returns the diff for a specific commit, showing what changes were made."
)
@openapi.tag("git")
@openapi.parameter(
    "file_path",
    description="Optional specific file to get diff for",
    _in="query",
    required=False,
    schema=str,
)
@openapi.response(
    200,
    {"application/json": {"commit": dict, "diff": str, "file_path": str}},
    description="Commit diff retrieved successfully",
)
@openapi.response(
    400,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Invalid commit SHA",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def handle_get_commit_diff(request: Request, commit_sha: str) -> HTTPResponse:
    """Handle get commit diff requests."""
    try:
        project_generator = get_project_generator(request)
        file_path = request.args.get("file_path")
        diff_data = await project_generator.git_service.get_commit_diff(
            commit_sha, file_path
        )
        return response.json(diff_data)
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.get_commit_diff.unexpected_error",
            tags={"endpoint": "/api/commits/<commit_sha>/diff"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to get commit diff",
                details={"error": str(exc)},
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@bp.route("/commits/<commit_sha>/revert", methods=["POST"])
@openapi.summary("Revert to commit")
@openapi.description(
    "Revert the last commit on the project. This creates a background job "
    "that adds a revert commit and retrains the agent. "
    "Returns immediately with a job ID. Connect to `/job-events/<job_id>` "
    "to receive server-sent events (SSE) for real-time progress tracking.\n\n"
    "**SSE Event Flow** (via `/job-events/<job_id>`):\n"
    "1. `received` - Request received by server\n"
    "2. `reverting` - Reverting last commit\n"
    "3. `revert_success` - Checkout revert commit\n"
    "4. `training` - Training the bot model with the reverted files\n"
    "5. `train_success` - Model training completed\n"
    "6. `done` - Revert completed\n\n"
    "**Error Events:**\n"
    "- `revert_error` - Failed to revert last commit\n"
    "- `train_error` - Revert succeeded but training failed\n"
    "- `error` - Unexpected error occurred\n\n"
    "**Usage:**\n"
    "1. Send POST request\n"
    "2. The response will be a JSON object `{job_id: ...}`\n"
    "3. Connect to `/job-events/<job_id>` for a server-sent event stream of progress."
)
@openapi.tag("git")
@openapi.response(
    200,
    {"application/json": model_to_schema(JobCreateResponse)},
    description="Job created. Poll or subscribe to /job-events/<job_id> for progress.",
)
@openapi.response(
    400,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Invalid commit SHA",
)
@openapi.response(
    500,
    {"application/json": model_to_schema(ApiErrorResponse)},
    description="Internal server error",
)
@openapi.parameter(
    HEADER_USER_ID,
    description=(
        "Optional user id to associate requests (e.g., for telemetry/guardrails)."
    ),
    _in="header",
    required=False,
    schema=str,
)
async def handle_revert_to_commit(request: Request, commit_sha: str) -> HTTPResponse:
    """Handle revert to commit requests."""
    try:
        # Validate commit SHA format
        if not commit_sha or len(commit_sha) < 7:
            return response.json(
                ApiErrorResponse(
                    error="Invalid commit SHA", details={"commit_sha": commit_sha}
                ).model_dump(),
                status=400,
            )

        job = job_manager.create_job()
        request.app.add_task(run_revert_job(request.app, job, commit_sha))
        return response.json(JobCreateResponse(job_id=job.id).model_dump(), status=200)
    except Exception as exc:
        capture_exception_with_context(
            exc,
            "bot_builder_service.revert_to_commit.unexpected_error",
            tags={"endpoint": "/api/commits/<commit_sha>/revert"},
        )
        return response.json(
            ApiErrorResponse(
                error="Failed to create revert job",
                details={"error": str(exc)},
            ).model_dump(),
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )


@langfuse.observe(capture_input=False, capture_output=False)
async def get_tracker_context_for_copilot(
    request: Request,
    req: CopilotTurnRequest,
    user_id: str,
) -> Optional[TrackerContext]:
    """Check the assistant chat for guardrail policy violations.

    Args:
        request: The request object.
        req: The CopilotTurnRequest object.
        user_id: The user ID.

    Returns:
        The tracker context if the tracker is available.
    """
    tracker = await current_tracker_from_input_channel(request.app, req.session_id)
    tracker_context = TrackerContext.from_tracker(
        tracker, max_turns=COPILOT_ASSISTANT_TRACKER_MAX_TURNS
    )
    if (
        tracker_context is not None
        and llm_service.guardrails_policy_checker is not None
    ):
        tracker_context = await llm_service.guardrails_policy_checker.check_assistant_chat_for_policy_violations(  # noqa: E501
            tracker_context=tracker_context,
            hello_rasa_user_id=user_id,
            hello_rasa_project_id=HELLO_RASA_PROJECT_ID,
            lakera_project_id=LAKERA_ASSISTANT_HISTORY_GUARDRAIL_PROJECT_ID,
        )

    # Track the retrieved tracker context
    CopilotLangfuseTelemetry.trace_copilot_tracker_context(
        tracker_context=tracker_context,
        max_conversation_turns=COPILOT_ASSISTANT_TRACKER_MAX_TURNS,
        session_id=req.session_id,
    )

    return tracker_context


@langfuse.observe(capture_input=False, capture_output=False)
def get_relevant_assistant_files_for_copilot(
    project_generator: ProjectGenerator,
) -> BotFiles:
    """Get the relevant assistant files for the copilot.

    Args:
        project_generator: The project generator.

    Returns:
        The relevant assistant files.
    """
    # Copilot doesn't need to know about the docs and any file that is not a core
    # assistant file
    files = project_generator.get_bot_files(
        exclude_docs_directory=True,
        allowed_file_extensions=["yaml", "yml", "py", "jinja", "jinja2"],
    )

    # Track the retrieved assistant files
    CopilotLangfuseTelemetry.trace_copilot_relevant_assistant_files(
        relevant_assistant_files=files,
    )
    return files



import os

# The configuration for the evaluator module.
CLAIM_EXTRACTOR_MODEL = os.getenv("CLAIM_EXTRACTOR_MODEL", "gpt-4.1-2025-04-14")
CLAIM_EXTRACTOR_TEMPERATURE = float(os.getenv("CLAIM_EXTRACTOR_TEMPERATURE", "0"))
CLAIM_EXTRACTOR_TIMEOUT = int(os.getenv("CLAIM_EXTRACTOR_TIMEOUT", "100"))
CLAIM_EXTRACTOR_MAX_TOKENS = int(os.getenv("CLAIM_EXTRACTOR_MAX_TOKENS", "2000"))



"""Authentication utilities for bot builder service."""

from datetime import datetime, timedelta, timezone
from functools import wraps
from typing import Any, Awaitable, Callable, Dict, Optional, Tuple

import jwt
import structlog
from jwt import PyJWKClient
from pydantic import BaseModel
from sanic import response
from sanic.request import Request
from sanic.response import HTTPResponse

from rasa.builder.config import (
    AUTH0_CLIENT_ID,
    AUTH0_ISSUER,
    AUTH_REQUIRED_AFTER_MINUTES,
    JWKS_URL,
)
from rasa.builder.models import ApiErrorResponse
from rasa.builder.project_info import ProjectInfo

HEADER_USER_ID = "X-User-Id"

structlogger = structlog.get_logger()

RouteHandler = Callable[..., Awaitable[HTTPResponse]]
ProtectedDecorator = Callable[[RouteHandler], RouteHandler]


def verify_auth0_token(token: str) -> Dict[str, Any]:
    """Verify JWT token from Auth0.

    Args:
        token: JWT token string

    Returns:
        Dict containing the token payload

    Raises:
        Exception: If token verification fails
    """
    jwk_client = PyJWKClient(JWKS_URL)
    signing_key = jwk_client.get_signing_key_from_jwt(token)

    payload = jwt.decode(
        token,
        signing_key.key,
        algorithms=["RS256"],
        audience=AUTH0_CLIENT_ID,
        issuer=AUTH0_ISSUER,
    )
    return payload


class Auth0TokenVerificationResult(BaseModel):
    payload: Optional[Dict[str, Any]]
    error_message: Optional[str]


def extract_and_verify_auth0_token(
    auth_header: str,
) -> Auth0TokenVerificationResult:
    """Extract and verify JWT token from Authorization header.

    Args:
        auth_header: Authorization header value

    Returns:
        Auth0TokenVerificationResult: Contains payload and error_message.
    """
    # Check Authorization header format
    if not auth_header.startswith("Bearer "):
        return Auth0TokenVerificationResult(
            payload=None, error_message="Missing or invalid Authorization header"
        )

    # Extract token
    token = auth_header.split(" ")[1]

    # Verify token
    try:
        payload = verify_auth0_token(token)
        return Auth0TokenVerificationResult(payload=payload, error_message=None)
    except Exception as e:
        return Auth0TokenVerificationResult(
            payload=None, error_message=f"Invalid token: {e!s}"
        )


def is_auth_required_now(
    *, always_required: bool = False, project_info: Optional[ProjectInfo] = None
) -> bool:
    """Return whether authentication is required right now based on config.

    - If always_required is True, returns True
    - Else, returns True iff current UTC time >= first_used + configured delta
    """
    if always_required:
        return True
    now = datetime.now(timezone.utc)

    if project_info is None:
        return False

    first_used = project_info.first_used_dt()
    if first_used is None:
        return False

    delta = timedelta(minutes=AUTH_REQUIRED_AFTER_MINUTES)
    auth_required_start = (first_used + delta).astimezone(timezone.utc)
    return now >= auth_required_start


def _check_and_attach_auth(request: Any) -> Tuple[bool, Optional[str]]:
    """Validate Authorization header. On success, store payload on request.ctx."""
    auth_header = request.headers.get("Authorization", "")
    verification_result = extract_and_verify_auth0_token(auth_header)
    if verification_result.error_message is not None:
        return False, verification_result.error_message

    # Attach payload for downstream consumers
    try:
        request.ctx.auth_payload = verification_result.payload
    except Exception:
        # If ctx is not available for some reason, ignore silently
        pass
    return True, None


def protected(*, always_required: bool = False) -> ProtectedDecorator:
    """Decorator to protect routes with conditional authentication.

    Authentication rules:
    - If always_required=True: token is always required
    - Else: token is required only after RASA_AUTH_AFTER timestamp

    On failure returns a 401 with a JSON error body.
    """

    def decorator(f: RouteHandler) -> RouteHandler:
        @wraps(f)
        async def decorated_function(
            request: Request, *args: Any, **kwargs: Any
        ) -> HTTPResponse:
            # Global bypass: allow turning off auth entirely
            if not getattr(request.app.config, "USE_AUTHENTICATION", True):
                return await f(request, *args, **kwargs)

            project_info: Optional[ProjectInfo] = None
            try:
                project_info = request.app.ctx.project_generator.project_info
            except Exception:
                project_info = None

            if is_auth_required_now(
                always_required=always_required, project_info=project_info
            ):
                ok, err = _check_and_attach_auth(request)
                if not ok:
                    structlogger.error(
                        "builder.auth.token_verification_failed", error=err
                    )
                    return response.json(
                        ApiErrorResponse(
                            error=err,
                            details={"expected": "Bearer <valid_token>"},
                        ).model_dump(),
                        status=401,
                    )
            return await f(request, *args, **kwargs)

        return decorated_function

    return decorator


def email_from_auth(request: Request) -> Optional[str]:
    """Retrieve user email from authentication information."""
    return (getattr(request.ctx, "auth_payload", None) or {}).get("sub")



"""Utilities for agent authentication and secret validation."""

import re
from typing import Any, Dict

from rasa.exceptions import ValidationError
from rasa.shared.agents.auth.constants import (
    CONFIG_API_KEY_KEY,
    CONFIG_CLIENT_SECRET_KEY,
    CONFIG_OAUTH_KEY,
    CONFIG_TOKEN_KEY,
)
from rasa.shared.constants import SECRET_DATA_FORMAT_PATTERN

AUTH_SECRETS = [CONFIG_API_KEY_KEY, CONFIG_TOKEN_KEY, CONFIG_CLIENT_SECRET_KEY]


def _is_valid_secret_data_format(value: str) -> bool:
    """Check if a value is in the correct environment variable format.

    Args:
        value: The value to check

    Returns:
        True if the value is in format "${env_var}", False otherwise
    """
    if not isinstance(value, str):
        return False

    # Use the common regex pattern for environment variable validation
    return bool(re.match(SECRET_DATA_FORMAT_PATTERN, value))


def _validate_secret_value(value: Any, key: str, context: str) -> None:
    """Generic function to validate a single secret value.

    Args:
        value: The value to validate
        key: The key name for error messages
        context: Context for error messages
            (e.g., "agent 'my_agent'", "MCP server 'my_server'")
    """
    if isinstance(value, str):
        if not _is_valid_secret_data_format(value):
            raise ValidationError(
                code="validation.sensitive_key_string_value_must_be_set_as_env_var",
                event_info=(
                    f"You defined the '{key}' in {context} as a string. The '{key}' "
                    f"must be set as an environment variable. Please update your "
                    f"config."
                ),
                key=key,
            )
    else:
        raise ValidationError(
            code="validation.sensitive_key_must_be_set_as_env_var",
            event_info=(
                f"You should define the '{key}' in {context} using the environment "
                f"variable syntax - ${{ENV_VARIABLE_NAME}}. Please update your config."
            ),
            key=key,
        )


def validate_secrets_in_params(
    params: Dict[str, Any], context_name: str = "configuration"
) -> None:
    """Validate that secrets in params are in environment variable format.

    Args:
        params: The parameters dictionary to validate
        context_name: Name of the context for error messages
            (e.g., "agent", "MCP server")
    """
    for key, value in params.items():
        if key in AUTH_SECRETS:
            _validate_secret_value(value, key, context_name)
        elif key == CONFIG_OAUTH_KEY and isinstance(value, dict):
            # Handle oauth object specifically - we know it contains client_secret
            if CONFIG_CLIENT_SECRET_KEY in value:
                _validate_secret_value(
                    value[CONFIG_CLIENT_SECRET_KEY],
                    CONFIG_CLIENT_SECRET_KEY,
                    context_name,
                )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import structlog

import rasa.dialogue_understanding.stack.utils as utils
from rasa.dialogue_understanding.commands.command import Command
from rasa.dialogue_understanding.patterns.correction import (
    FLOW_PATTERN_CORRECTION_ID,
    CorrectionPatternFlowStackFrame,
)
from rasa.dialogue_understanding.stack.dialogue_stack import DialogueStack
from rasa.dialogue_understanding.stack.frames.flow_stack_frame import (
    BaseFlowStackFrame,
)
from rasa.shared.core.events import Event
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.flows.flow_step import FlowStepWithFlowReference
from rasa.shared.core.flows.steps.constants import END_STEP
from rasa.shared.core.flows.steps.continuation import ContinueFlowStep
from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger()


@dataclass
class CorrectedSlot:
    """A slot that was corrected."""

    name: str
    value: Any
    filled_by: Optional[str] = None


@dataclass
class CorrectSlotsCommand(Command):
    """A command to correct the value of a slot."""

    corrected_slots: List[CorrectedSlot]

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        return "correct slot"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> CorrectSlotsCommand:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        try:
            return CorrectSlotsCommand(
                corrected_slots=[
                    CorrectedSlot(
                        s["name"], value=s["value"], filled_by=s.get("filled_by", None)
                    )
                    for s in data["corrected_slots"]
                ]
            )
        except KeyError as e:
            raise ValueError(
                f"Missing key when parsing CorrectSlotsCommand: {e}"
            ) from e

    @staticmethod
    def are_all_slots_reset_only(
        proposed_slots: Dict[str, Any], all_flows: FlowsList
    ) -> bool:
        """Checks if all slots are reset only.

        A slot is reset only if the `collect` step it gets filled by
        has the `ask_before_filling` flag set to `True`. This means, the slot
        shouldn't be filled if the question isn't asked.

        If such a slot gets corrected, we don't want to correct the slot but
        instead reset the flow to the question where the slot was asked.

        Args:
            proposed_slots: The proposed slots.
            all_flows: All flows in the assistant.

        Returns:
            `True` if all slots are reset only, `False` otherwise.
        """
        return all(
            collect_step.collect not in proposed_slots
            or collect_step.ask_before_filling
            for flow in all_flows.underlying_flows
            for collect_step in flow.get_collect_steps()
        )

    @staticmethod
    def find_earliest_updated_collect_info(
        updated_slots: Dict[str, Any],
        all_flows: FlowsList,
        tracker: DialogueStateTracker,
    ) -> Optional[FlowStepWithFlowReference]:
        """Find the earliest collect information step that fills one of the slots.

        When we update slots, we need to reset a flow to the question when the slot
        was asked. This function finds the earliest collect information step that
        fills one of the slots - with the idea being that we afterwards go through
        the other updated slots.

        Args:
            updated_slots: The slots that were updated.
            all_flows: All flows.
            tracker: The dialogue state tracker.

        Returns:
        The earliest collect information step that fills one of the slots and
        the  flow id of that step.
        """

        collect_steps = utils.previous_collect_steps_for_active_flow(tracker, all_flows)

        for collect_step, flow_id in collect_steps:
            if collect_step.collect in updated_slots:
                return FlowStepWithFlowReference(collect_step, flow_id)

        return None

    def corrected_slots_dict(self, tracker: DialogueStateTracker) -> Dict[str, Any]:
        """Returns the slots that should be corrected.

        Filters out slots, that are already set to the correct value.

        Args:
            tracker: The tracker.

        Returns:
        A dict with the slots and their values that should be corrected.
        """
        proposed_slots = {}
        for corrected_slot in self.corrected_slots:
            if tracker.get_slot(corrected_slot.name) != corrected_slot.value:
                proposed_slots[corrected_slot.name] = {
                    "value": corrected_slot.value,
                    "filled_by": corrected_slot.filled_by,
                }
            else:
                structlogger.debug(
                    "correct_slots_command.skip_correction.slot_already_set",
                    command=self,
                )
        return proposed_slots

    @staticmethod
    def determine_index_for_new_correction_frame(
        top_flow_frame: BaseFlowStackFrame, stack: DialogueStack
    ) -> int:
        """Returns the index for the correction frame.

        Args:
            top_flow_frame: The top flow frame.
            stack: The stack.

        Returns:
            The index for the correction frame.
        """
        if top_flow_frame.flow_id != FLOW_PATTERN_CORRECTION_ID:
            # we are not in a correction flow, so we can just push the correction
            # frame on top of the stack
            return len(stack.frames)
        else:
            # we allow the previous correction to finish first before
            # starting the new one. that's why we insert the new correction below
            # the previous one.
            for i, frame in enumerate(stack.frames):
                if frame.frame_id == top_flow_frame.frame_id:
                    return i
            else:
                # we should never get here as we should always find the previous
                # correction frame
                raise ValueError(
                    f"Could not find the previous correction frame "
                    f"{top_flow_frame.frame_id} on the stack {stack}."
                )

    @staticmethod
    def end_previous_correction(
        top_flow_frame: BaseFlowStackFrame, stack: DialogueStack
    ) -> None:
        """Ends the previous correction.

        If the top flow frame is already a correction, we wrap up the previous
        correction before starting the new one. All frames that were added
        after that correction and the correction itself will be set to continue
        at the END step.

        Args:
            top_flow_frame: The top flow frame.
            stack: The stack.
        """
        if top_flow_frame.flow_id != FLOW_PATTERN_CORRECTION_ID:
            # only need to end something if we are already in a correction
            return

        for frame in reversed(stack.frames):
            if isinstance(frame, BaseFlowStackFrame):
                frame.step_id = ContinueFlowStep.continue_step_for_id(END_STEP)
                if frame.frame_id == top_flow_frame.frame_id:
                    break

    @classmethod
    def create_correction_frame(
        cls,
        proposed_slots: Dict[str, Any],
        all_flows: FlowsList,
        tracker: DialogueStateTracker,
    ) -> Optional[CorrectionPatternFlowStackFrame]:
        """Creates a correction frame.

        Args:
            proposed_slots: The proposed slots.
            all_flows: All flows in the assistant.
            tracker: The dialogue state tracker.

        Returns:
            The correction frame.
        """
        # check if all corrected slots have ask_before_filling=True
        # if this is a case, we are not correcting a value but we
        # are resetting the slots and jumping back to the first question
        is_reset_only = cls.are_all_slots_reset_only(proposed_slots, all_flows)

        earliest_collect = cls.find_earliest_updated_collect_info(
            proposed_slots, all_flows, tracker
        )

        return CorrectionPatternFlowStackFrame(
            is_reset_only=is_reset_only,
            corrected_slots=proposed_slots,
            reset_flow_id=earliest_collect.flow_id if earliest_collect else None,
            reset_step_id=earliest_collect.step.id if earliest_collect else None,
            new_slot_values=[
                value.get("value") for slot, value in proposed_slots.items()
            ],
        )

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        """Runs the command on the tracker.

        Args:
            tracker: The tracker to run the command on.
            all_flows: All flows in the assistant.
            original_tracker: The tracker before any command was executed.

        Returns:
            The events to apply to the tracker.
        """
        stack = tracker.stack

        top_flow_frame = utils.top_flow_frame(stack)
        if not top_flow_frame:
            # we shouldn't end up here as a correction shouldn't be triggered
            # if we are not in any flow. but just in case we do, we
            # just skip the command.
            structlogger.warning("correct_slots_command.no_active_flow")
            return []

        structlogger.debug("correct_slots_command", command=self)

        # check if the correct slot is referring to a slot of a flow on the stack
        # the slot also needs to be part of a collect step in any of those flows
        # if this is not the case, we don't want to correct the slot
        for slot in self.corrected_slots:
            if not self.should_correct_slot(slot, tracker, all_flows):
                structlogger.warning(
                    "correct_slots_command.skip_correct_slot",
                    correct_slot=slot,
                    reason="The slot is not part of a collect step in any of the flows "
                    "on the stack. Skipping correction.",
                )
                return []

        proposed_slots = self.corrected_slots_dict(tracker)

        correction_frame = self.create_correction_frame(
            proposed_slots, all_flows, tracker
        )
        if not correction_frame:
            return []

        insertion_index = self.determine_index_for_new_correction_frame(
            top_flow_frame, stack
        )
        self.end_previous_correction(top_flow_frame, stack)

        stack.push(correction_frame, index=insertion_index)
        return tracker.create_stack_updated_events(stack)

    def __hash__(self) -> int:
        return hash(self.command())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CorrectSlotsCommand):
            return False

        return True

    def should_correct_slot(
        self, slot: CorrectedSlot, tracker: DialogueStateTracker, all_flows: FlowsList
    ) -> bool:
        """Checks if the slot should be corrected.

        Args:
            slot: The slot to check.
            tracker: The tracker.
            all_flows: All flows in the assistant.
        """
        # get all flows on the stack
        flows_on_stack = utils.user_flows_on_the_stack(tracker.stack)

        # check if the slot is part of a collect step in any of the flows on the stack
        for flow_id in flows_on_stack:
            flow = all_flows.flow_by_id(flow_id)
            if flow is None:
                continue
            for collect_step in flow.get_collect_steps():
                if collect_step.collect == slot.name:
                    return True

        return False



import rasa.e2e_test.constants as constants

KEY_FIXTURES = constants.KEY_FIXTURES
KEY_METADATA = constants.KEY_METADATA
KEY_STUB_CUSTOM_ACTIONS = constants.KEY_STUB_CUSTOM_ACTIONS
KEY_USER_INPUT = constants.KEY_USER_INPUT
KEY_BOT_INPUT = constants.KEY_BOT_INPUT
KEY_BOT_UTTERED = constants.KEY_BOT_UTTERED
KEY_STEPS = constants.KEY_STEPS
KEY_TEST_CASE = constants.KEY_TEST_CASE
KEY_TEST_CASES = constants.KEY_TEST_CASES
KEY_COMMANDS = "commands"
KEY_ADDITIONAL_COMMANDS = "additional_commands"
KEY_REMOVE_DEFAULT_COMMANDS = "remove_default_commands"

ACTOR_USER = "user"
ACTOR_BOT = "bot"

DEFAULT_INPUT_TESTS_PATH = "dialogue_understanding_tests/"
SCHEMA_FILE_PATH = "dialogue_understanding_test/du_test_schema.yml"

PLACEHOLDER_GENERATED_ANSWER_TEMPLATE = "placeholder_generated_answer"



"""High-level authentication management with strategy abstraction."""

from typing import Any, Dict, Optional

import structlog

from rasa.shared.agents.auth.agent_auth_factory import AgentAuthFactory
from rasa.shared.agents.auth.auth_strategy import AgentAuthStrategy
from rasa.shared.agents.auth.constants import (
    CONFIG_API_KEY_KEY,
    CONFIG_MODULE_KEY,
    CONFIG_OAUTH_KEY,
    CONFIG_TOKEN_KEY,
)
from rasa.shared.agents.auth.types import AgentAuthType
from rasa.shared.exceptions import AgentAuthInitializationException

structlogger = structlog.get_logger()


class AgentAuthManager:
    """High-level authentication management with strategy abstraction."""

    def __init__(self, auth_strategy: Optional[AgentAuthStrategy] = None):
        """Initialize a new authentication manager instance."""
        self._auth_strategy: Optional[AgentAuthStrategy] = auth_strategy

    def get_auth(self) -> AgentAuthStrategy:
        """Retrieve the authentication instance.

        Returns:
            The authentication instance.

        Raises:
            ValueError: If no authentication is connected.
        """
        if self._auth_strategy is None:
            raise ValueError("No authentication instance available")
        return self._auth_strategy

    @staticmethod
    def detect_auth_type(config: Dict[str, Any]) -> AgentAuthType:
        if CONFIG_MODULE_KEY in config:
            return AgentAuthType.CUSTOM
        if CONFIG_API_KEY_KEY in config:
            return AgentAuthType.API_KEY
        if CONFIG_TOKEN_KEY in config:
            return AgentAuthType.BEARER_TOKEN
        if CONFIG_OAUTH_KEY in config:
            return AgentAuthType.OAUTH2
        raise ValueError(
            "Invalid authentication type. Supported types: "
            + ", ".join([auth_type.value for auth_type in AgentAuthType])
        )

    @classmethod
    def load_auth(
        cls, config: Optional[Dict[str, Any]]
    ) -> Optional["AgentAuthManager"]:
        """Connect to authentication using specified strategy type and persist
        the auth instance to the manager in a ready-to-use state.

        Args:
            config: The configuration dictionary for the authentication.

        Raises:
            AgentAuthInitializationException: If the authentication connection fails.
        """
        if not config:
            return None
        try:
            auth_type = AgentAuthManager.detect_auth_type(config)

            # Create the auth client
            client = AgentAuthFactory.create_client(auth_type, config)

            structlogger.debug(
                "agent_auth_manager.load_auth.success",
                auth_type=auth_type.value,
                event_info=(
                    f"Loaded authentication client successfully for `{auth_type.value}`"
                ),
            )
            return cls(client)
        except Exception as e:
            event_info = "Failed to load authentication client"
            structlogger.error(
                "agent_auth_manager.load_auth.failed_to_load",
                event_info=event_info,
                config=config,
            )
            raise AgentAuthInitializationException(e) from e



from typing import Optional

from mcp.types import CallToolResult
from pydantic import BaseModel


class AgentToolResult(BaseModel):
    tool_name: str
    result: Optional[str] = None
    is_error: bool = False
    error_message: Optional[str] = None

    def __str__(self) -> str:
        """Provide a clean string representation for logging."""
        if self.is_error:
            return (
                f"AgentToolResult(tool_name='{self.tool_name}', "
                f"error='{self.error_message}')"
            )
        return f"AgentToolResult(tool_name='{self.tool_name}', result={self.result})"

    def __repr__(self) -> str:
        """Provide a detailed representation for debugging."""
        return self.__str__()

    @classmethod
    def from_mcp_tool_result(
        cls, tool_name: str, tool_result: CallToolResult
    ) -> "AgentToolResult":
        if tool_result.isError:
            return cls(
                tool_name=tool_name,
                result=None,
                is_error=tool_result.isError,
                error_message=str(tool_result.content),
            )

        # try to use structured content if available
        if tool_result.structuredContent:
            if "result" in tool_result.structuredContent:
                return cls(
                    tool_name=tool_name,
                    result=str(tool_result.structuredContent["result"]),
                    is_error=tool_result.isError,
                )
            return cls(
                tool_name=tool_name,
                result=str(tool_result.structuredContent),
                is_error=tool_result.isError,
            )
        # fallback to content if structured content is not available
        elif tool_result.content:
            return cls(
                tool_name=tool_name,
                result=str([content.model_dump() for content in tool_result.content]),
                is_error=tool_result.isError,
            )
        # if no content is available, return None
        else:
            return cls(
                tool_name=tool_name,
                result=None,
                is_error=tool_result.isError,
                error_message="No content returned from tool `{tool_name}`",
            )



from rasa.agents.schemas.agent_input import AgentInput, AgentInputSlot
from rasa.agents.schemas.agent_output import AgentOutput
from rasa.agents.schemas.agent_tool_result import AgentToolResult
from rasa.agents.schemas.agent_tool_schema import AgentToolSchema, CustomToolSchema

__all__ = [
    "AgentInput",
    "AgentInputSlot",
    "AgentOutput",
    "AgentToolSchema",
    "AgentToolResult",
    "CustomToolSchema",
]



import logging
from typing import Any, Dict, Optional, Text

import aiohttp
import structlog

from rasa.core import constants
from rasa.core.channels import UserMessage
from rasa.shared.nlu.constants import INTENT_NAME_KEY
from rasa.utils.endpoints import EndpointConfig

logger = logging.getLogger(__name__)
structlogger = structlog.get_logger()


class RasaNLUHttpInterpreter:
    """Allows for an HTTP endpoint to be used to parse messages."""

    def __init__(self, endpoint_config: Optional[EndpointConfig] = None) -> None:
        """Initializes a `RasaNLUHttpInterpreter`."""
        self.session = aiohttp.ClientSession()
        if endpoint_config:
            self.endpoint_config = endpoint_config
        else:
            self.endpoint_config = EndpointConfig(constants.DEFAULT_SERVER_URL)

    async def parse(self, message: UserMessage) -> Dict[Text, Any]:
        """Parse a text message.

        Return a default value if the parsing of the text failed.
        """
        default_return = {
            "intent": {INTENT_NAME_KEY: "", "confidence": 0.0},
            "entities": [],
            "text": "",
        }

        result = await self._rasa_http_parse(message.text, message.sender_id)
        return result if result is not None else default_return

    async def _rasa_http_parse(
        self, text: Text, message_id: Optional[Text] = None
    ) -> Optional[Dict[Text, Any]]:
        """Send a text message to a running rasa NLU http server.

        Return `None` on failure.
        """
        if not self.endpoint_config or self.endpoint_config.url is None:
            structlogger.error(
                "http.parse.text",
                event_info="No rasa NLU server specified!",
            )
            return None

        params = {
            "token": self.endpoint_config.token,
            "text": text,
            "message_id": message_id,
        }

        if self.endpoint_config.url.endswith("/"):
            url = self.endpoint_config.url + "model/parse"
        else:
            url = self.endpoint_config.url + "/model/parse"

        # noinspection PyBroadException
        try:
            async with self.session.post(url, json=params) as resp:
                if resp.status == 200:
                    return await resp.json()
                else:
                    structlogger.error(
                        "http.parse.text.failure",
                        event_info="Failed to parse text",
                    )
                    return None
        except Exception as e:  # skipcq: PYL-W0703
            # need to catch all possible exceptions when doing http requests
            # (timeouts, value errors, parser errors, ...)
            structlogger.exception(
                "http.parse.text.exception",
                event_info=f"Exception occurred while parsing text. Error: {e}",
            )
            return None



from dataclasses import dataclass

from rasa.dialogue_understanding.commands.command import Command


@dataclass
class FreeFormAnswerCommand(Command):
    """A command to indicate a free-form answer by the bot."""

    pass



import logging
from asyncio import CancelledError
from typing import Any, Awaitable, Callable, Dict, Iterable, List, Optional, Text, Union

import cachecontrol
import google.auth.transport.requests
import requests
import structlog
from google.oauth2 import id_token
from sanic import Blueprint, response
from sanic.exceptions import SanicException
from sanic.request import Request
from sanic.response import HTTPResponse

from rasa.core.channels.channel import InputChannel, OutputChannel, UserMessage

logger = logging.getLogger(__name__)
structlogger = structlog.get_logger()

CHANNEL_NAME = "hangouts"
CERTS_URL = (
    "https://www.googleapis.com/service_accounts/"
    "v1/metadata/x509/chat@system.gserviceaccount.com"
)


class HangoutsOutput(OutputChannel):
    """A Hangouts communication channel."""

    @classmethod
    def name(cls) -> Text:
        """Return channel name."""
        return CHANNEL_NAME

    def __init__(self) -> None:
        """Starts messages as empty dictionary."""
        self.messages: Dict[Text, Any] = {}

    @staticmethod
    def _text_card(message: Dict[Text, Any]) -> Dict:
        card = {
            "cards": [
                {
                    "sections": [
                        {"widgets": [{"textParagraph": {"text": message["text"]}}]}
                    ]
                }
            ]
        }
        return card

    @staticmethod
    def _image_card(image: Text) -> Dict:
        card = {
            "cards": [{"sections": [{"widgets": [{"image": {"imageUrl": image}}]}]}]
        }
        return card

    @staticmethod
    def _text_button_card(text: Text, buttons: List) -> Union[Dict, None]:
        hangouts_buttons = []
        for b in buttons:
            try:
                b_txt, b_pl = b["title"], b["payload"]
            except KeyError:
                logger.error(
                    "Buttons must be a list of dicts with 'title' and 'payload' as keys"
                )
                return None

            hangouts_buttons.append(
                {
                    "textButton": {
                        "text": b_txt,
                        "onClick": {"action": {"actionMethodName": b_pl}},
                    }
                }
            )

        card = {
            "cards": [
                {
                    "sections": [
                        {
                            "widgets": [
                                {"textParagraph": {"text": text}},
                                {"buttons": hangouts_buttons},
                            ]
                        }
                    ]
                }
            ]
        }
        return card

    @staticmethod
    def _combine_cards(c1: Dict, c2: Dict) -> Dict:
        return {"cards": [*c1["cards"], *c2["cards"]]}

    async def _persist_message(self, message: Dict) -> None:
        """Google Hangouts only accepts single dict with single key 'text'
        for simple text messages. All other responses must be sent as cards.

        In case the bot sends multiple messages, all are transformed to either
        cards or text output
        """
        # check whether current and previous message will send 'text' or 'card'
        if self.messages.get("text"):
            msg_state = "text"
        elif self.messages.get("cards"):
            msg_state = "cards"
        else:
            msg_state = None

        if message.get("text"):
            msg_new = "text"
        elif message.get("cards"):
            msg_new = "cards"
        else:
            raise Exception(
                "Your message to Hangouts channel must either contain 'text' or "
                "'cards'!"
            )

        # depending on above outcome, convert messages into same type and combine
        if msg_new == msg_state == "text":
            # two text messages are simply appended
            new_text = " ".join([self.messages.get("text", ""), message["text"]])
            new_messages = {"text": new_text}

        elif msg_new == msg_state == "cards":
            # two cards are combined into one
            new_messages = self._combine_cards(self.messages, message)

        elif msg_state == "cards" and msg_new == "text":
            # if any message is card, turn text message into TextParagraph card
            # and combine cards
            text_card = self._text_card(message)
            new_messages = self._combine_cards(self.messages, text_card)

        elif msg_state == "text" and msg_new == "cards":
            text_card = self._text_card(self.messages)
            new_messages = self._combine_cards(text_card, message)

        elif msg_new == "text":
            new_messages = {"text": message["text"]}
        else:
            new_messages = message

        self.messages = new_messages

    async def send_text_message(
        self, recipient_id: Text, text: Text, **kwargs: Any
    ) -> None:
        await self._persist_message({"text": text})

    async def send_image_url(
        self, recipient_id: Text, image: Text, **kwargs: Any
    ) -> None:
        await self._persist_message(self._image_card(image))

    async def send_text_with_buttons(
        self, recipient_id: Text, text: Text, buttons: List, **kwargs: Any
    ) -> None:
        await self._persist_message(self._text_button_card(text, buttons))

    async def send_attachment(
        self, recipient_id: Text, attachment: Text, **kwargs: Any
    ) -> None:
        await self.send_text_message(recipient_id, attachment)

    async def send_elements(
        self, recipient_id: Text, elements: Iterable[Dict[Text, Any]], **kwargs: Any
    ) -> None:
        raise NotImplementedError

    async def send_custom_json(
        self, recipient_id: Text, json_message: Dict, **kwargs: Any
    ) -> None:
        """Custom json payload is simply forwarded to Google Hangouts without
        any modifications. Use this for more complex cards, which can be created
        in actions.py.
        """
        await self._persist_message(json_message)


# Google Hangouts input channel
class HangoutsInput(InputChannel):
    """Channel that uses Google Hangouts Chat API to communicate."""

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> InputChannel:
        if credentials:
            return cls(credentials.get("project_id"))

        return cls()

    def __init__(
        self,
        project_id: Optional[Text] = None,
        hangouts_user_added_intent_name: Optional[Text] = "/user_added",
        hangouts_room_added_intent_name: Optional[Text] = "/room_added",
        hangouts_removed_intent_name: Optional[Text] = "/bot_removed",
    ) -> None:
        self.project_id = project_id
        self.hangouts_user_added_intent_name = hangouts_user_added_intent_name
        self.hangouts_room_added_intent_name = hangouts_room_added_intent_name
        self.hangouts_user_added_intent_name = hangouts_removed_intent_name

        # Google's Request obj (this is used to make HTTP requests) uses cached
        # session to fetch Google's service certs. Certs don't change frequently,
        # so it makes sense to cache request body, rather than getting it again
        # every message. Actual caching depends on response headers.
        # see: https://github.com/googleapis/google-auth-library-python/blob/main/google/oauth2/id_token.py#L15 # noqa: E501
        cached_session = cachecontrol.CacheControl(requests.session())
        self.google_request = google.auth.transport.requests.Request(  # type: ignore[no-untyped-call]
            session=cached_session
        )

    @classmethod
    def name(cls) -> Text:
        """Returns channel name."""
        return CHANNEL_NAME

    @staticmethod
    def _extract_sender(req: Request) -> Text:
        if req.json["type"] == "MESSAGE":
            return req.json["message"]["sender"]["displayName"]

        return req.json["user"]["displayName"]

    # noinspection PyMethodMayBeStatic
    def _extract_message(self, req: Request) -> Text:
        if req.json["type"] == "MESSAGE":
            message = req.json["message"]["text"]

        elif req.json["type"] == "CARD_CLICKED":
            message = req.json["action"]["actionMethodName"]

        elif req.json["type"] == "ADDED_TO_SPACE":
            if self._extract_room(req) and self.hangouts_room_added_intent_name:
                message = self.hangouts_room_added_intent_name
            elif not self._extract_room(req) and self.hangouts_user_added_intent_name:
                message = self.hangouts_user_added_intent_name

        elif (
            req.json["type"] == "REMOVED_FROM_SPACE"
            and self.hangouts_user_added_intent_name
        ):
            message = self.hangouts_user_added_intent_name
        else:
            message = ""

        return message

    @staticmethod
    def _extract_room(req: Request) -> Union[Text, None]:
        if req.json["space"]["type"] == "ROOM":
            return req.json["space"]["displayName"]

        return None

    def _extract_input_channel(self) -> Text:
        return self.name()

    def _check_token(self, bot_token: Text) -> None:
        # see https://developers.google.com/chat/how-tos/bots-develop#verifying_bot_authenticity # noqa: E501
        # and https://google-auth.readthedocs.io/en/latest/user-guide.html#identity-tokens # noqa: E501
        try:
            decoded_token = id_token.verify_token(  # type: ignore[no-untyped-call]
                bot_token,
                self.google_request,
                audience=self.project_id,
                certs_url=CERTS_URL,
            )
        except ValueError:
            raise SanicException(status_code=401)
        if decoded_token["iss"] != "chat@system.gserviceaccount.com":
            raise SanicException(status_code=401)

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[None]]
    ) -> Blueprint:
        """API configuration for the channel webhook."""
        custom_webhook = Blueprint("hangouts_webhook", __name__)

        @custom_webhook.route("/", methods=["GET"])
        async def health(request: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @custom_webhook.route("/webhook", methods=["POST"])
        async def receive(request: Request) -> HTTPResponse:
            if self.project_id:
                token = request.headers.get("Authorization", "").replace("Bearer ", "")
                self._check_token(token)

            sender_id = self._extract_sender(request)
            room_name = self._extract_room(request)
            text = self._extract_message(request)
            if text is None:
                return response.text("OK")
            input_channel = self._extract_input_channel()

            collector = HangoutsOutput()

            try:
                await on_new_message(
                    UserMessage(
                        text,
                        collector,
                        sender_id,
                        input_channel=input_channel,
                        metadata={"room": room_name},
                    )
                )
            except CancelledError as e:
                structlogger.error(
                    "hangouts.message.blueprint.timeout",
                    event_info=f"Message processing was cancelled. Error: {e}",
                )
            except Exception as e:
                structlogger.exception(
                    "hangouts.message.blueprint.failure",
                    event_info=(
                        f"An error occurred while processing the message. Error: {e}",
                    ),
                )

            return response.json(collector.messages)

        return custom_webhook



import os
from typing import Any, Dict, Optional, Set

import boto3
import structlog
from botocore.exceptions import BotoCoreError, ClientError
from google.auth.environment_vars import AWS_DEFAULT_REGION

from rasa.shared.constants import (
    API_BASE_CONFIG_KEY,
    API_VERSION_CONFIG_KEY,
    AWS_ACCESS_KEY_ID_CONFIG_KEY,
    AWS_ACCESS_KEY_ID_ENV_VAR,
    AWS_BEDROCK_PROVIDER,
    AWS_REGION_NAME_CONFIG_KEY,
    AWS_REGION_NAME_ENV_VAR,
    AWS_SAGEMAKER_CHAT_PROVIDER,
    AWS_SAGEMAKER_PROVIDER,
    AWS_SECRET_ACCESS_KEY_CONFIG_KEY,
    AWS_SECRET_ACCESS_KEY_ENV_VAR,
    AWS_SESSION_TOKEN_CONFIG_KEY,
    AZURE_API_BASE_ENV_VAR,
    AZURE_API_VERSION_ENV_VAR,
    DEPLOYMENT_CONFIG_KEY,
)
from rasa.shared.exceptions import ProviderClientValidationError
from rasa.shared.utils.io import resolve_environment_variables

structlogger = structlog.get_logger()


def validate_aws_setup_for_litellm_clients(
    litellm_model_name: str, litellm_call_kwargs: Dict, source_log: str, provider: str
) -> None:
    """Validates the AWS setup for LiteLLM clients to ensure credentials are set.

    Args:
        litellm_model_name (str): The name of the LiteLLM model being validated.
        litellm_call_kwargs (dict): Additional keyword arguments passed to the client,
            which may include configuration values for AWS credentials.
        source_log (str): The source log identifier for structured logging.
        provider (str): The provider for which the validation is being performed.

    Raises:
        ProviderClientValidationError: If any required AWS environment variable
            or corresponding configuration key is missing.
    """
    from rasa.shared.utils.health_check.health_check import is_api_health_check_enabled

    # expand environment variables if referenced in the config
    resolved_litellm_call_kwargs: Dict = resolve_environment_variables(
        litellm_call_kwargs
    )  # type: ignore[assignment]

    # boto3 only accepts bedrock and sagemaker as valid clients
    # therefore we need to convert the provider name if it is defined
    # as sagemaker_chat
    provider = (
        AWS_SAGEMAKER_PROVIDER if provider == AWS_SAGEMAKER_CHAT_PROVIDER else provider
    )

    # if the AWS credentials are defined in the endpoints yaml model config,
    # either as referenced secret env vars or direct values, we need to pass them
    # to the boto3 client to ensure that the client can connect to the AWS service.
    additional_kwargs: Dict[str, Any] = {}
    if AWS_ACCESS_KEY_ID_CONFIG_KEY in resolved_litellm_call_kwargs:
        additional_kwargs[AWS_ACCESS_KEY_ID_CONFIG_KEY] = resolved_litellm_call_kwargs[
            AWS_ACCESS_KEY_ID_CONFIG_KEY
        ]
    if AWS_SECRET_ACCESS_KEY_CONFIG_KEY in resolved_litellm_call_kwargs:
        additional_kwargs[AWS_SECRET_ACCESS_KEY_CONFIG_KEY] = (
            resolved_litellm_call_kwargs[AWS_SECRET_ACCESS_KEY_CONFIG_KEY]
        )
    if AWS_SESSION_TOKEN_CONFIG_KEY in resolved_litellm_call_kwargs:
        additional_kwargs[AWS_SESSION_TOKEN_CONFIG_KEY] = resolved_litellm_call_kwargs[
            AWS_SESSION_TOKEN_CONFIG_KEY
        ]
    if AWS_REGION_NAME_CONFIG_KEY in resolved_litellm_call_kwargs:
        additional_kwargs["region_name"] = resolved_litellm_call_kwargs[
            AWS_REGION_NAME_CONFIG_KEY
        ]

    if is_api_health_check_enabled():
        structlogger.debug(
            f"{source_log}.validating_aws_credentials_for_litellm_clients_via_aws_client",
            model_name=litellm_model_name,
        )
        _validate_credentials_with_aws_client(
            provider,
            additional_kwargs,
            litellm_model_name,
            source_log,
        )
        return None

    return _validate_credentials_exist(
        resolved_litellm_call_kwargs, litellm_model_name, source_log
    )


def _validate_credentials_exist(
    resolved_litellm_call_kwargs: Dict[str, Any],
    litellm_model_name: str,
    source_log: str,
) -> None:
    """Validates that AWS credentials are provided.

    Args:
        resolved_litellm_call_kwargs (Dict[str, Any]): The resolved keyword arguments
            containing AWS credentials.
        litellm_model_name (str): The name of the LiteLLM model being validated.
        source_log (str): The source log identifier for structured logging.

    Raises:
        ProviderClientValidationError: If any required AWS credentials are missing.
    """
    required_iam_credential_keys = {"model_id", AWS_REGION_NAME_CONFIG_KEY}
    required_env_var_secrets = {
        AWS_ACCESS_KEY_ID_CONFIG_KEY,
        AWS_SECRET_ACCESS_KEY_CONFIG_KEY,
        AWS_REGION_NAME_CONFIG_KEY,
    }

    provided_credentials: Dict[str, Any] = _add_env_vars_credentials_if_defined(
        resolved_litellm_call_kwargs
    )
    provided_credentials_keys = set(provided_credentials.keys())
    common_credentials = (
        required_iam_credential_keys | required_env_var_secrets
    ) & provided_credentials_keys

    if not common_credentials:
        event_info = (
            "Missing AWS credentials for LiteLLM clients. "
            "Ensure that you are using one of the available authentication methods: "
            "endpoints yml, environment variables, or IAM roles. "
        )
        structlogger.error(
            f"{source_log}.validate_aws_credentials_existence_for_litellm_clients",
            event_info=event_info,
            model_name=litellm_model_name,
        )
        raise ProviderClientValidationError(event_info)

    # if the chosen auth method is via env vars,
    # ensure that all required env vars are set
    if common_credentials.issubset(required_env_var_secrets):
        _verify_missing_credentials(
            required_env_var_secrets, common_credentials, litellm_model_name, source_log
        )

    if common_credentials.issubset(required_iam_credential_keys):
        _verify_missing_credentials(
            required_iam_credential_keys,
            common_credentials,
            litellm_model_name,
            source_log,
        )


def _add_env_vars_credentials_if_defined(
    resolved_litellm_call_kwargs: Dict[str, Any],
) -> Dict[str, Any]:
    """Adds AWS credentials from environment variables to the resolved call kwargs.

    Args:
        resolved_litellm_call_kwargs (Dict[str, Any]): The resolved keyword arguments.

    Returns:
        Dict[str, Any]: The updated keyword arguments with AWS credentials from
            environment variables if they were defined.
    """
    if AWS_ACCESS_KEY_ID_CONFIG_KEY not in resolved_litellm_call_kwargs:
        env_var_value = os.getenv(AWS_ACCESS_KEY_ID_ENV_VAR)
        if env_var_value is not None:
            resolved_litellm_call_kwargs[AWS_ACCESS_KEY_ID_CONFIG_KEY] = env_var_value

    if AWS_SECRET_ACCESS_KEY_CONFIG_KEY not in resolved_litellm_call_kwargs:
        env_var_value = os.getenv(AWS_SECRET_ACCESS_KEY_ENV_VAR)
        if env_var_value is not None:
            resolved_litellm_call_kwargs[AWS_SECRET_ACCESS_KEY_CONFIG_KEY] = (
                env_var_value
            )

    if AWS_REGION_NAME_CONFIG_KEY not in resolved_litellm_call_kwargs:
        env_var_value = os.getenv(
            AWS_REGION_NAME_ENV_VAR, os.getenv(AWS_DEFAULT_REGION)
        )
        if env_var_value is not None:
            resolved_litellm_call_kwargs[AWS_REGION_NAME_CONFIG_KEY] = env_var_value

    return resolved_litellm_call_kwargs


def _verify_missing_credentials(
    required_credentials: Set[str],
    present_credentials: Set[str],
    litellm_model_name: str,
    source_log: str,
) -> None:
    """Verifies if any required credentials are missing.

    Args:
        present_credentials (set): The set of essential credential keys
            that are present.
        required_credentials (set): The set of required credential keys.
        litellm_model_name (str): The name of the LiteLLM model being validated.
        source_log (str): The source log identifier for structured logging.

    Raises:
        ProviderClientValidationError: If any required credentials are missing.
    """
    missing_credentials = required_credentials - present_credentials
    if missing_credentials:
        event_info = (
            f"Missing AWS credentials for "
            f"LiteLLM clients: {', '.join(missing_credentials)}. "
        )
        structlogger.error(
            f"{source_log}.validate_aws_credentials_existence_for_litellm_clients",
            event_info=event_info,
            model_name=litellm_model_name,
        )
        raise ProviderClientValidationError(event_info)


def _validate_credentials_with_aws_client(
    provider: str,
    additional_kwargs: Dict[str, Any],
    litellm_model_name: str,
    source_log: str,
) -> None:
    """Creates an AWS client with the provided credentials.

    Args:
        provider (str): The AWS service provider.
        additional_kwargs (Dict[str, Any]): Additional keyword arguments for the client.
        litellm_model_name (str): The name of the LiteLLM model being validated.
        source_log (str): The source log identifier for structured logging.
    """
    try:
        # We are using the boto3 client because it can discover the AWS credentials
        # from the environment variables, credentials file, or IAM roles.
        # This is necessary to ensure that the client can connect to the AWS service.
        aws_client = boto3.client(provider, **additional_kwargs)

        # Using different method calls available to different AWS clients
        # to test the connection
        if provider == AWS_SAGEMAKER_PROVIDER:
            aws_client.list_models()
        elif provider == AWS_BEDROCK_PROVIDER:
            aws_client.get_model_invocation_logging_configuration()

    except (ClientError, BotoCoreError) as exc:
        event_info = (
            f"Failed to validate AWS setup for LiteLLM clients: {exc}. "
            f"Ensure that you are using one of the available authentication methods:"
            f"credentials file, environment variables, or IAM roles. "
            f"Also, ensure that the AWS region is set correctly. "
        )
        structlogger.error(
            f"{source_log}.validate_aws_credentials_for_litellm_clients_via_aws_client.failed",
            event_info=event_info,
            exception=str(exc),
            model_name=litellm_model_name,
        )
        raise ProviderClientValidationError(event_info)


def validate_azure_client_setup(
    api_base: Optional[str],
    api_version: Optional[str],
    deployment: Optional[str],
) -> None:
    """Validates the Azure setup for LiteLLM Router clients to ensure
     that all required configuration parameters are set.
    Raises:
        ProviderClientValidationError: If any required Azure configurations
            is missing.
    """

    def generate_event_info_for_missing_setting(
        setting: str,
        setting_env_var: Optional[str] = None,
        setting_config_key: Optional[str] = None,
    ) -> str:
        """Generate a part of the message with instructions on what to set
        for the missing client setting.
        """
        info = "Set {setting} with {options}. "
        options = ""
        if setting_env_var is not None:
            options += f"environment variable '{setting_env_var}'"
        if setting_config_key is not None and setting_env_var is not None:
            options += " or "
        if setting_config_key is not None:
            options += f"config key '{setting_config_key}'"

        return info.format(setting=setting, options=options)

    # All required settings for Azure OpenAI client
    settings: Dict[str, Dict[str, Any]] = {
        "API Base": {
            "current_value": api_base,
            "env_var": AZURE_API_BASE_ENV_VAR,
            "config_key": API_BASE_CONFIG_KEY,
        },
        "API Version": {
            "current_value": api_version,
            "env_var": AZURE_API_VERSION_ENV_VAR,
            "config_key": API_VERSION_CONFIG_KEY,
        },
        "Deployment Name": {
            "current_value": deployment,
            "env_var": None,
            "config_key": DEPLOYMENT_CONFIG_KEY,
        },
    }

    missing_settings = [
        setting_name
        for setting_name, setting_info in settings.items()
        if setting_info["current_value"] is None
    ]

    if missing_settings:
        event_info = f"Client settings not set: {', '.join(missing_settings)}. "

        for missing_setting in missing_settings:
            if settings[missing_setting]["current_value"] is not None:
                continue
            event_info += generate_event_info_for_missing_setting(
                missing_setting,
                settings[missing_setting]["env_var"],
                settings[missing_setting]["config_key"],
            )

        structlogger.error(
            "azure_openai_llm_client.not_configured",
            event_info=event_info,
            missing_settings=missing_settings,
        )
        raise ProviderClientValidationError(event_info)



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Dict, Iterable, Optional, Text

import structlog
from boto3.dynamodb.conditions import Key

import rasa.utils
from rasa.constants import DEFAULT_SANIC_WORKERS, ENV_SANIC_WORKERS
from rasa.core.tracker_stores.tracker_store import (
    SerializedTrackerAsDict,
    TrackerStore,
)
from rasa.shared.core.domain import Domain
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import RasaException
from rasa.utils.endpoints import EndpointConfig

structlogger = structlog.get_logger(__name__)

if TYPE_CHECKING:
    import boto3.resources.factory.dynamodb.Table


class DynamoTrackerStore(TrackerStore, SerializedTrackerAsDict):
    """Stores conversation history in DynamoDB."""

    def __init__(
        self,
        domain: Domain,
        table_name: Text = "states",
        region: Text = "us-east-1",
        event_broker: Optional[EndpointConfig] = None,
        **kwargs: Dict[Text, Any],
    ) -> None:
        """Initialize `DynamoTrackerStore`.

        Args:
            domain: Domain associated with this tracker store.
            table_name: The name of the DynamoDB table, does not need to be present a
                priori.
            region: The name of the region associated with the client.
                A client is associated with a single region.
            event_broker: An event broker used to publish events.
            kwargs: Additional kwargs.
        """
        import boto3

        self.client = boto3.client("dynamodb", region_name=region)
        self.region = region
        self.table_name = table_name
        self.db = self.get_or_create_table(table_name)
        super().__init__(domain, event_broker, **kwargs)

    def get_or_create_table(
        self, table_name: Text
    ) -> "boto3.resources.factory.dynamodb.Table":
        """Returns table or creates one if the table name is not in the table list."""
        import boto3

        dynamo = boto3.resource("dynamodb", region_name=self.region)
        try:
            self.client.describe_table(TableName=table_name)
        except self.client.exceptions.ResourceNotFoundException:
            sanic_workers_count = int(
                os.environ.get(ENV_SANIC_WORKERS, DEFAULT_SANIC_WORKERS)
            )

            if sanic_workers_count > 1:
                structlogger.error(
                    "dynamo_tracker_store.table_creation_not_supported_in_multi_worker_mode",
                    event_info=(
                        "DynamoDB table creation is not "
                        "supported in multi-worker mode. "
                        "Table should already exist.",
                    ),
                )
                raise RasaException(
                    "DynamoDB table creation is not supported in "
                    "case of multiple sanic workers. To create the table either "
                    "run Rasa with a single worker or create the table manually."
                    "Here are the defaults which can be used to "
                    "create the table manually: "
                    f"Table name: {table_name}, Primary key: sender_id, "
                    f"key type `HASH`, attribute type `S` (String), "
                    "Provisioned throughput: Read capacity units: 5, "
                    "Write capacity units: 5"
                )

            table = dynamo.create_table(
                TableName=self.table_name,
                KeySchema=[{"AttributeName": "sender_id", "KeyType": "HASH"}],
                AttributeDefinitions=[
                    {"AttributeName": "sender_id", "AttributeType": "S"}
                ],
                ProvisionedThroughput={"ReadCapacityUnits": 5, "WriteCapacityUnits": 5},
            )

            # Wait until the table exists.
            table.meta.client.get_waiter("table_exists").wait(TableName=table_name)
        else:
            table = dynamo.Table(table_name)

        return table

    async def save(self, tracker: DialogueStateTracker) -> None:
        """Saves the current conversation state."""
        await self.stream_events(tracker)
        serialized = self.serialise_tracker(tracker)

        full_tracker = await self.retrieve_full_tracker(tracker.sender_id)
        if full_tracker is None:
            self.db.put_item(Item=serialized)
            return None

        # return the latest events since the last user message
        new_tracker = DialogueStateTracker.from_dict(
            serialized["sender_id"], events_as_dict=serialized["events"]
        )
        new_events = new_tracker.get_last_turn_events()
        new_serialized_events = [event.as_dict() for event in new_events]

        # we need to save the full tracker if it is a new tracker
        # without events following a user message
        if not new_serialized_events:
            self.db.put_item(Item=serialized)
            return None

        # append new events to the existing tracker
        self.db.update_item(
            Key={"sender_id": tracker.sender_id},
            UpdateExpression="SET events = list_append(if_not_exists(events, :empty_list), :events)",  # noqa: E501
            ExpressionAttributeValues={
                ":events": new_serialized_events,
                ":empty_list": [],
            },
            ReturnValues="UPDATED_NEW",
        )
        return None

    async def delete(self, sender_id: Text) -> None:
        """Delete tracker for the given sender_id."""
        if not await self.exists(sender_id):
            structlogger.info(
                "dynamo_tracker_store.delete.no_tracker_for_sender_id",
                event_info=f"Could not find tracker for conversation ID '{sender_id}'.",
            )
            return None

        self.db.delete_item(
            Key={"sender_id": sender_id},
            ConditionExpression="attribute_exists(sender_id)",
        )
        structlogger.info(
            "dynamo_tracker_store.delete.deleted_tracker",
            sender_id=sender_id,
        )

    @staticmethod
    def serialise_tracker(
        tracker: "DialogueStateTracker",
    ) -> Dict:
        """Serializes the tracker, returns object with decimal types.

        DynamoDB cannot store `float`s, so we'll convert them to `Decimal`s.
        """
        return rasa.utils.json_utils.replace_floats_with_decimals(
            SerializedTrackerAsDict.serialise_tracker(tracker)
        )

    async def retrieve(self, sender_id: Text) -> Optional[DialogueStateTracker]:
        """Retrieve dialogues for a sender_id in reverse-chronological order.

        Based on the session_date sort key.
        """
        return await self._retrieve(sender_id, fetch_all_sessions=False)

    async def retrieve_full_tracker(
        self, sender_id: Text
    ) -> Optional[DialogueStateTracker]:
        """Retrieves tracker for all conversation sessions.

        Args:
            sender_id: Conversation ID to fetch the tracker for.
        """
        return await self._retrieve(sender_id, fetch_all_sessions=True)

    async def _retrieve(
        self, sender_id: Text, fetch_all_sessions: bool
    ) -> Optional[DialogueStateTracker]:
        """Returns tracker matching sender_id.

        Args:
            sender_id: Conversation ID to fetch the tracker for.
            fetch_all_sessions: Whether to fetch all sessions or only the last one.
        """
        dialogues = self.db.query(
            KeyConditionExpression=Key("sender_id").eq(sender_id),
            ScanIndexForward=False,
        )["Items"]

        if not dialogues:
            return None

        events_with_floats = []
        for dialogue in dialogues:
            if dialogue.get("events"):
                events = rasa.utils.json_utils.replace_decimals_with_floats(
                    dialogue["events"]
                )
                events_with_floats.extend(events)

        if self.domain is None:
            slots = []
        else:
            slots = self.domain.slots

        tracker = DialogueStateTracker.from_dict(sender_id, events_with_floats, slots)

        if fetch_all_sessions:
            return tracker

        # only return the last session
        multiple_tracker_sessions = (
            rasa.shared.core.trackers.get_trackers_for_conversation_sessions(tracker)
        )

        if len(multiple_tracker_sessions) <= 1:
            return tracker

        return multiple_tracker_sessions[-1]

    async def keys(self) -> Iterable[Text]:
        """Returns sender_ids of the `DynamoTrackerStore`."""
        response = self.db.scan(ProjectionExpression="sender_id")
        sender_ids = [i["sender_id"] for i in response["Items"]]

        while response.get("LastEvaluatedKey"):
            response = self.db.scan(
                ProjectionExpression="sender_id",
                ExclusiveStartKey=response["LastEvaluatedKey"],
            )
            sender_ids.extend([i["sender_id"] for i in response["Items"]])

        return sender_ids

    async def update(
        self, tracker: DialogueStateTracker, apply_deletion_only: bool = True
    ) -> None:
        """Overwrites the tracker for the given sender_id."""
        serialized = self.serialise_tracker(tracker)
        self.db.put_item(Item=serialized)

        structlogger.info(
            "dynamo_tracker_store.replace.replaced_tracker",
            sender_id=tracker.sender_id,
        )



import argparse
import os
from typing import TYPE_CHECKING, Optional

import structlog

from rasa import telemetry
from rasa.core.constants import DEFAULT_SUB_AGENTS
from rasa.exceptions import ValidationError
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.shared.utils.common import display_research_study_prompt

if TYPE_CHECKING:
    from rasa.validator import Validator

structlogger = structlog.get_logger()

FREE_TEXT_INPUT_PROMPT = "Type out your own message..."


def _validate_domain(validator: "Validator") -> bool:
    valid_domain_validity = validator.verify_domain_validity()
    valid_actions_in_stories_rules = validator.verify_actions_in_stories_rules()
    valid_forms_in_stories_rules = validator.verify_forms_in_stories_rules()
    valid_form_slots = validator.verify_form_slots()
    valid_slot_mappings = validator.verify_slot_mappings()
    valid_responses = validator.check_for_no_empty_parenthesis_in_responses()
    valid_buttons = validator.validate_button_payloads()
    valid_slot_validation = validator.verify_slot_validation()
    valid_conditional_responses = (
        validator.validate_conditional_response_variation_predicates()
    )
    return (
        valid_domain_validity
        and valid_actions_in_stories_rules
        and valid_forms_in_stories_rules
        and valid_form_slots
        and valid_slot_mappings
        and valid_responses
        and valid_buttons
        and valid_slot_validation
        and valid_conditional_responses
    )


def _validate_nlu(validator: "Validator", fail_on_warnings: bool) -> bool:
    return validator.verify_nlu(not fail_on_warnings)


def _validate_story_structure(
    validator: "Validator", max_history: Optional[int], fail_on_warnings: bool
) -> bool:
    # Check if a valid setting for `max_history` was given
    if isinstance(max_history, int) and max_history < 1:
        raise argparse.ArgumentTypeError(
            f"The value of `--max-history {max_history}` is not a positive integer."
        )

    return validator.verify_story_structure(
        not fail_on_warnings, max_history=max_history
    )


def _validate_sub_agents(sub_agents_path: str) -> bool:
    """Validates sub-agents configuration.

    Args:
        sub_agents_path: Path to the sub-agents directory.

    Returns:
        True if validation passes, False otherwise.
    """
    from rasa.agents.validation import validate_agent_folder
    from rasa.core.constants import DEFAULT_SUB_AGENTS

    try:
        # Check if the sub-agents directory exists
        if not os.path.isdir(sub_agents_path):
            # If the sub-agents-path points to the default folder and it doesn't exist,
            # no agents are available.
            # AvailableAgents will handle the non-existing folder gracefully.

            if sub_agents_path == DEFAULT_SUB_AGENTS:
                structlogger.info(
                    "cli.validate_files.sub_agents_validation",
                    sub_agents_path=sub_agents_path,
                    event_info="Default sub-agents directory does not exist, "
                    "no sub-agents will be available.",
                )
                return True
            else:
                # For user-specified paths, the directory must exist
                structlogger.error(
                    "cli.validate_files.sub_agents_validation_error",
                    sub_agents_path=sub_agents_path,
                    event_info=f"Sub-agents directory '{sub_agents_path}' "
                    "does not exist.",
                )
                return False

        # Validate the actual config content using AvailableAgents
        # This will validate file existence, structure, mandatory keys, etc.
        try:
            validate_agent_folder(sub_agents_path)

        except ValidationError as e:
            # This is a validation error - log it and return False
            structlogger.error(
                "cli.validate_files.sub_agents_validation_error",
                sub_agents_path=sub_agents_path,
                validation_error=str(e),
                event_info=f"Sub-agents configuration validation failed: {e}",
            )
            return False

        except Exception as e:
            # This is an unexpected error
            structlogger.error(
                "cli.validate_files.sub_agents_validation_error",
                sub_agents_path=sub_agents_path,
                error=str(e),
                event_info=f"Unexpected error during sub-agents validation: {e}",
            )
            return False

        structlogger.info(
            "cli.validate_files.sub_agents_validation_success",
            sub_agents_path=sub_agents_path,
            event_info="Sub-agents validation passed successfully.",
        )
        return True

    except Exception as e:
        # This is an unexpected error
        structlogger.error(
            "cli.validate_files.sub_agents_validation_error",
            sub_agents_path=sub_agents_path,
            error=str(e),
            event_info="Sub-agents validation failed.",
        )
        return False


def validate_files(
    fail_on_warnings: bool,
    max_history: Optional[int],
    importer: TrainingDataImporter,
    stories_only: bool = False,
    flows_only: bool = False,
    translations_only: bool = False,
    sub_agents: Optional[str] = None,
    endpoints: Optional[str] = None,
) -> None:
    """Validates either the story structure or the entire project.

    Args:
        fail_on_warnings: `True` if the process should exit with a non-zero status
        max_history: The max history to use when validating the story structure.
        importer: The `TrainingDataImporter` to use to load the training data.
        stories_only: If `True`, only the story structure is validated.
        flows_only: If `True`, only the flows are validated.
        translations_only: If `True`, only the translations data is validated.
        sub_agents: Path to sub-agents directory for validation.
        endpoints: Path to the endpoints configuration file.
    """
    from rasa.validator import Validator

    validator = Validator.from_importer(importer)

    if stories_only:
        all_good = _validate_story_structure(validator, max_history, fail_on_warnings)
    elif flows_only:
        all_good = validator.verify_flows()
    elif translations_only:
        all_good = validator.verify_translations()
    else:
        # Check for empty user-defined domain
        # before proceeding with other validations.
        # The user domain could be empty if the
        # provided domain is missing or invalid
        if importer.get_user_domain().is_empty():
            display_research_study_prompt()
            raise ValidationError(
                code="cli.validate_files.empty_domain",
                event_info="Encountered empty domain during validation.",
            )

        valid_domain = _validate_domain(validator)
        valid_nlu = _validate_nlu(validator, fail_on_warnings)
        valid_stories = _validate_story_structure(
            validator, max_history, fail_on_warnings
        )
        valid_flows = validator.verify_flows()
        valid_prompt_templates = validator.verify_prompt_templates()
        if validator.config:
            valid_translations = validator.verify_translations(summary_mode=True)
        else:
            valid_translations = True
        valid_CALM_slot_mappings = validator.validate_CALM_slot_mappings()

        # Validate sub-agents
        sub_agents_path = sub_agents or DEFAULT_SUB_AGENTS
        valid_sub_agents = _validate_sub_agents(sub_agents_path)

        if valid_sub_agents:
            valid_sub_agents = validator.validate_agent_flow_conflicts(sub_agents_path)

        all_good = (
            valid_domain
            and valid_nlu
            and valid_stories
            and valid_flows
            and valid_prompt_templates
            and valid_translations
            and valid_CALM_slot_mappings
            and valid_sub_agents
        )

    if validator.config:
        validator.warn_if_config_mandatory_keys_are_not_set()

    telemetry.track_validate_files(all_good)
    if not all_good:
        display_research_study_prompt()
        raise ValidationError(
            code="cli.validate_files.project_validation_error",
            event_info="Project validation completed with errors.",
        )
